import warnings
import os

warnings.simplefilter(action="ignore", category=FutureWarning)


# Ensure NumPy only uses 1 thread for matrix multiplication,
# because numpy is stupid and tries to use heaps of threads which is quite wasteful
# and it makes our models run way more slowly.
os.environ["OMP_NUM_THREADS"] = "1"



"""
Functions needed for dealing with age stratification
"""


def add_zero_to_age_breakpoints(breakpoints):
    """
    append a zero on to a list if there isn't one already present, for the purposes of age stratification

    :param breakpoints: list
        integers for the age breakpoints requested
    :return: list
        age breakpoints with the zero value included
    """
    return [0] + breakpoints if 0 not in breakpoints else breakpoints


def split_age_parameter(age_breakpoints, parameter):
    """
    creates a dictionary to request splitting of a parameter according to age breakpoints, but using values of 1 for
        each age stratum
    allows that later parameters that might be age-specific can be modified for some age strata

    :param age_breakpoints: list
        list of the age breakpoints to be requested, with breakpoints as string
    :param parameter: str
        name of parameter that will need to be split
    :return: dict
        dictionary with age groups as string as keys and ones for all the values
    """
    age_breakpoints = ["0"] + age_breakpoints if "0" not in age_breakpoints else age_breakpoints
    return {parameter: {str(age_group): 1.0 for age_group in age_breakpoints}}



from typing import List, Tuple, Dict, Callable

import numpy as np

from summer.legacy.constants import Flow as FlowType
from summer.legacy.compartment import Compartment
from summer.legacy.stratification import Stratification

from .base import BaseEntryFlow


class ImportFlow(BaseEntryFlow):
    """
    Calculates importation, based on the current population size.
    """

    type = FlowType.IMPORT

    def __init__(
        self,
        dest: Compartment,
        param_name: str,
        param_func: Callable[[str, float], float],
        adjustments: list = [],
    ):
        assert type(dest) is Compartment
        self.adjustments = adjustments
        self.dest = dest
        self.param_name = param_name
        self.param_func = param_func

    def get_net_flow(self, compartment_values, time):
        return self.get_weight_value(time)

    def copy(self, **kwargs):
        """
        Creates a modified copy of the flow for stratification.
        """
        return ImportFlow(
            dest=kwargs["dest"],
            param_name=kwargs["param_name"],
            param_func=kwargs["param_func"],
            adjustments=kwargs["adjustments"],
        )

    def __repr__(self):
        return f"<ImportFlow to {self.dest} with {self.param_name}>"



from typing import List, Tuple, Dict, Callable

import numpy as np

from .age_stratification import add_zero_to_age_breakpoints, split_age_parameter
from .data_structures import (
    convert_boolean_list_to_indices,
    create_cumulative_dict,
    element_list_division,
    element_list_multiplication,
    increment_list_by_index,
    normalise_dict,
    order_dict_by_keys,
)
from .stratification_funcs import (
    create_additive_function,
    create_function_of_function,
    create_multiplicative_function,
    create_sloping_step_function,
    create_time_variant_multiplicative_function,
)
from .validation import validate_stratify, validate_model



from typing import List, Tuple, Dict, Callable
from summer.legacy.constants import Flow as FlowType
from summer.legacy.compartment import Compartment
from summer.legacy.stratification import Stratification

from .base import BaseExitFlow


class BaseDeathFlow(BaseExitFlow):
    """A flow representing deaths"""

    def __init__(
        self,
        source: Compartment,
        param_name: str,
        param_func: Callable[[str, float], float],
        adjustments: list = [],
    ):
        assert type(source) is Compartment
        self.adjustments = adjustments
        self.source = source
        self.param_name = param_name
        self.param_func = param_func

    def get_net_flow(self, compartment_values, time):
        parameter_value = self.get_weight_value(time)
        population = compartment_values[self.source.idx]
        flow_rate = parameter_value * population
        return flow_rate

    def copy(self, **kwargs):
        """
        Creates a modified copy of the flow for stratification.
        """
        return self.__class__(
            source=kwargs["source"],
            param_name=kwargs["param_name"],
            param_func=kwargs["param_func"],
            adjustments=kwargs["adjustments"],
        )


class InfectionDeathFlow(BaseDeathFlow):
    """A flow representing infection deaths"""

    type = FlowType.DEATH

    def __repr__(self):
        return f"<DeathFlow from {self.source} with {self.param_name}>"


class UniversalDeathFlow(BaseDeathFlow):
    """A flow representing non-infection deaths"""

    type = FlowType.UNIVERSAL_DEATH

    def __repr__(self):
        return f"<UniversalDeathFlow from {self.source} with {self.param_name}>"



from .crude_birth import CrudeBirthFlow
from .infection import InfectionDensityFlow, InfectionFrequencyFlow, BaseInfectionFlow
from .replace_birth import ReplacementBirthFlow
from .standard import StandardFlow
from .death import InfectionDeathFlow, UniversalDeathFlow, BaseDeathFlow
from .ageing import AgeingFlow
from .importation import ImportFlow
from .base import BaseTransitionFlow, BaseExitFlow, BaseEntryFlow



import numpy
from scipy.integrate import quad
from summer.legacy.model import order_dict_by_keys, add_zero_to_age_breakpoints
from copy import copy


"""
pallettes of functions that may be useful for creating parameter values to submit to the SUMMER module
"""


def change_parameter_unit(parameter_dict, multiplier):
    """
    used to adapt the parameters according their unit - for example, could be used for models that are running in time
        steps that are different from the time step assumed by the input parameter

    :param parameter_dict: dict
        dictionary whose values need to be adjusted
    :param multiplier: float
        multiplier
    :return: dict
        dictionary with values multiplied by the multiplier argument
    """
    return {
        param_key: param_value * multiplier for param_key, param_value in parameter_dict.items()
    }


def add_w_to_param_names(parameter_dict):
    """
    add a "W" string to the end of the parameter name to indicate that the parameter should over-write up the chain of
        stratification, rather than being a multiplier or adjustment function for the upstream parameters

    :param parameter_dict: dict
        the dictionary before the adjustments
    :return: dict
        same dictionary but with the "W" string added to each of the keys
    """
    return {str(key) + "W": value for key, value in parameter_dict.items()}


def get_average_value_of_function(input_function, start_value, end_value):
    """
    use numeric integration to find the average value of a function between two extremes

    :param input_function: function
        function to be interrogated
    :param start_value: float
        lower limit of the independent variable over which to integrate the function
    :param end_value: float
        upper limit of the independent variable over which to integrate the function
    """
    return quad(input_function, start_value, end_value)[0] / (end_value - start_value)


def get_parameter_dict_from_function(input_function, breakpoints, upper_value=100.0):
    """
    create a dictionary of parameter values from a continuous function, an arbitrary upper value and some breakpoints
    within which to evaluate the function
    """
    revised_breakpoints = copy.copy(add_zero_to_age_breakpoints(breakpoints))
    revised_breakpoints.append(upper_value)
    param_values = []
    for n_breakpoint in range(len(revised_breakpoints) - 1):
        param_values.append(
            get_average_value_of_function(
                input_function,
                revised_breakpoints[n_breakpoint],
                revised_breakpoints[n_breakpoint + 1],
            )
        )
    return {str(key): value for key, value in zip(revised_breakpoints, param_values)}


def substratify_parameter(
    parameter_to_stratify,
    stratum_to_split,
    param_value_dict,
    breakpoints,
    stratification="age",
):
    """
    produce dictionary revise a stratum of a parameter that has been split at a higher level from dictionary of the
        values for each stratum of the higher level of the split

    :param parameter_to_stratify: str
        name of the parameter that was split at the higher level
    :param stratum_to_split: str
        stratum whose values should be revised
    :param param_value_dict: dict
        dictionary with keys age breakpoints and values parameter values
    :param breakpoints: list
        list of age breakpoints submitted as integer
    :param stratification: str
        name of the stratification this is being applied to
    :return: dict
        dictionary with keys being upstream stratified parameter to split and keys dictionaries with their keys the
            current stratum of interest and values the parameter multiplier
    """
    return {
        parameter_to_stratify
        + "X"
        + stratification
        + "_"
        + str(age_group): {stratum_to_split: param_value_dict[str(age_group)]}
        for age_group in add_zero_to_age_breakpoints(breakpoints)
    }


"""
functions that return a function of an independent variable
expected that this will often be a model quantity, such as time or age
"""


def get_parameter_dict_from_function(input_function, breakpoints, upper_value=100.0):
    """
    create a dictionary of parameter values from a continuous function, an arbitrary upper value and some breakpoints
        within which to evaluate the function
    """
    revised_breakpoints = copy(add_zero_to_age_breakpoints(breakpoints))
    revised_breakpoints.append(upper_value)
    param_values = []
    for n_breakpoint in range(len(revised_breakpoints) - 1):
        param_values.append(
            get_average_value_of_function(
                input_function,
                revised_breakpoints[n_breakpoint],
                revised_breakpoints[n_breakpoint + 1],
            )
        )
    return {str(key): value for key, value in zip(revised_breakpoints, param_values)}


def create_step_function_from_dict(input_dict):
    """
    create a step function out of dictionary with numeric keys and values, where the keys determine the values of the
        independent variable at which the steps between the output values occur

    :param input_dict: dict
        dictionary in standard format with numeric keys for the points at which the steps occur and numeric values for
            the values to be returned from these points onwards
    :return: function
        the function constructed from input data
    """
    dict_keys, dict_values = order_dict_by_keys(input_dict)

    def step_function(input_value):
        if input_value >= dict_keys[-1]:
            return dict_values[-1]
        else:
            for key in range(len(dict_keys)):
                if input_value < dict_keys[key + 1]:
                    return dict_values[key]

    return step_function


def sinusoidal_scaling_function(start_time, baseline_value, end_time, final_value):
    """
    in order to implement scale-up functions over time, use the cosine function to produce smooth scale-up functions
        from one point to another, returning the starting value before the starting point and the final value after the
        end point

    :param start_time: float
        starting value of the independent variable
    :param baseline_value: float
        starting value of the dependent variable
    :param end_time: float
        final value of the independent variable
    :param final_value: float
        final value of the dependent variable
    :return:
        function scaling from the starting value to the final value
    """

    def sinusoidal_function(time):
        if not isinstance(time, float):
            raise ValueError("value provided to scaling function not a float")
        elif start_time > end_time:
            raise ValueError("start time is later than end time")
        elif time < start_time:
            return baseline_value
        elif start_time <= time <= end_time:
            return baseline_value + (final_value - baseline_value) * (
                0.5 - 0.5 * numpy.cos((time - start_time) * numpy.pi / (end_time - start_time))
            )
        else:
            return final_value

    return sinusoidal_function


def logistic_scaling_function(parameter):
    """
    a specific sigmoidal form of function that scales up from zero to one around the point of the value of parameter

    :param parameter: float
        the single parameter to the function
    :return: function
        the logistic function
    """
    return lambda x: 1.0 - 1.0 / (1.0 + numpy.exp(-(parameter - x)))



"""
Functions to validate the inputs to a model.
Validation performed using Cerberus: https://docs.python-cerberus.org/en/stable/index.html
"""
from typing import List

from cerberus import Validator
import numpy as np


from summer.legacy.constants import (
    Compartment,
    Flow,
    BirthApproach,
    IntegrationType,
)


def validate_stratify(
    model,
    stratification_name,
    strata_request,
    compartments_to_stratify,
    requested_proportions,
    adjustment_requests,
    infectiousness_adjustments,
    mixing_matrix,
):
    schema = get_stratify_schema(
        model, stratification_name, strata_request, compartments_to_stratify
    )
    validator = Validator(schema, allow_unknown=True, require_all=True)
    stratify_data = {
        "stratification_name": stratification_name,
        "strata_request": strata_request,
        "compartments_to_stratify": compartments_to_stratify,
        "requested_proportions": requested_proportions,
        "adjustment_requests": adjustment_requests,
        "infectiousness_adjustments": infectiousness_adjustments,
        "mixing_matrix": mixing_matrix,
    }
    is_valid = validator.validate(stratify_data)
    if not is_valid:
        errors = validator.errors
        raise ValidationException(errors)


def get_stratify_schema(model, stratification_name, strata_names, compartments_to_stratify):
    """
    Schema used to validate model attributes during initialization.
    """
    prev_strat_names = [s.name for s in model.stratifications]
    strata_names_strs = [str(s) for s in strata_names]
    is_full_stratified = compartments_to_stratify == model.original_compartment_names
    return {
        "stratification_name": {
            "type": "string",
            "forbidden": prev_strat_names,
        },
        "strata_request": {
            "type": "list",
            "check_with": check_strata_request(
                stratification_name, compartments_to_stratify, model.original_compartment_names
            ),
            "schema": {"type": ["string", "integer"]},
        },
        "compartments_to_stratify": {
            "type": "list",
            "schema": {"type": "string"},
            "allowed": model.original_compartment_names,
        },
        "requested_proportions": {
            "type": "dict",
            "valuesrules": {"type": ["integer", "float"]},
            "keysrules": {"allowed": strata_names_strs},
        },
        "adjustment_requests": {
            "type": "dict",
            "keysrules": {
                # FIXME: Not quite right... yet.
                # "allowed": list(model.parameters.keys()),
                "check_with": check_adj_request_key(compartments_to_stratify),
            },
            "valuesrules": {
                "type": "dict",
                "keysrules": {"allowed": strata_names_strs + [f"{s}W" for s in strata_names_strs]},
                "valuesrules": {
                    "type": ["integer", "float", "string"],
                    "check_with": check_time_variant_key(model.time_variants),
                },
            },
        },
        "infectiousness_adjustments": {
            "type": "dict",
            "valuesrules": {"type": ["integer", "float"]},
            "keysrules": {"allowed": strata_names_strs},
        },
        "mixing_matrix": {
            "nullable": True,
            "check_with": check_mixing_matrix(strata_names, is_full_stratified),
        },
    }


def check_adj_request_key(compartments_to_stratify):
    """
    Check adjustment request keys
    """

    def _check(field, value, error):
        if not compartments_to_stratify and value == "universal_death_rate":
            error(
                field,
                "Universal death rate can only be adjusted for when all compartments are being stratified",
            )

    return _check


def check_strata_request(strat_name, compartments_to_stratify, model_compartments):
    """
    Strata requested must be well formed.
    """

    def _check(field, value, error):
        if strat_name == "age":
            if not min([int(s) for s in value]) == 0:
                error(field, "First age strata must be '0'")
            if compartments_to_stratify and not compartments_to_stratify == model_compartments:
                error(field, "Age stratification must be applied to all compartments")

    return _check


def check_time_variant_key(time_variants: dict):
    """
    Ensure value is a key in time variants if it is a string
    """

    def _check(field, value, error):
        if type(value) is str and value not in time_variants:
            error(field, "String value must be found in time variants dict.")

    return _check


def check_mixing_matrix(strata_names: List[str], is_full_stratified: bool):
    """
    Ensure mixing matrix is correctly specified
    """
    num_strata = len(strata_names)

    def _check(field, value, error):
        if value is None:
            return  # This is fine

        if not is_full_stratified:
            error(field, "Mixing matrix can only be applied to full stratifications.")

        if callable(value):
            # Dynamic mixing matrix
            mm = value(0)
        else:
            # Static mixing matrix
            mm = value

        if not type(mm) is np.ndarray:
            error(field, "Mixing matrix must be Numpy array (or None)")
        elif mm.shape != (num_strata, num_strata):
            error(field, f"Mixing matrix must have shape ({num_strata}, {num_strata})")

    return _check


def validate_model(model_kwargs):
    """
    Throws an error if the model's initial data is invalid.
    """
    starting_population = model_kwargs.get("starting_population")
    parameters = model_kwargs.get("parameters", {})
    compartment_names = model_kwargs.get("compartment_names", [])
    parameter_names = [*parameters.keys()]
    schema = get_model_schema(starting_population, parameter_names, compartment_names)
    validator = Validator(schema, allow_unknown=True, require_all=True)
    is_valid = validator.validate(model_kwargs)
    if not is_valid:
        errors = validator.errors
        raise ValidationException(errors)

    # Validate times seperately because Cerberus is slow.
    times = model_kwargs["times"]
    assert type(times) is np.ndarray, "Times must be a NumPy array."
    assert times.dtype == np.dtype("float64"), "Times must be float64."


def get_model_schema(starting_population, parameter_names, compartment_names):
    """
    Schema used to validate model attributes during initialization.
    """
    return {
        "starting_population": {"type": "integer"},
        "entry_compartment": {"type": "string", "allowed": compartment_names},
        "birth_approach": {
            "type": "string",
            "allowed": [
                BirthApproach.ADD_CRUDE,
                BirthApproach.REPLACE_DEATHS,
                BirthApproach.NO_BIRTH,
            ],
        },
        "compartment_names": {"type": "list", "schema": {"type": "string"}},
        "infectious_compartments": {
            "type": "list",
            "schema": {"type": "string"},
            "allowed": compartment_names,
        },
        "initial_conditions": {
            "type": "dict",
            "valuesrules": {"anyof_type": ["integer", "float"]},
            "check_with": check_initial_conditions(starting_population, compartment_names),
        },
        "requested_flows": {
            "type": "list",
            "check_with": check_requested_flows(parameter_names, compartment_names),
            "schema": {
                "type": "dict",
                "schema": {
                    "type": {
                        "type": "string",
                        "allowed": [
                            Flow.STANDARD,
                            Flow.INFECTION_FREQUENCY,
                            Flow.INFECTION_DENSITY,
                            Flow.DEATH,
                            Flow.IMPORT,
                        ],
                    },
                    "parameter": {"type": "string"},
                    "origin": {"type": "string", "required": False},
                    "to": {"type": "string", "required": False},
                },
            },
        },
    }


def check_requested_flows(parameter_names, compartment_names):
    """
    Validate flows
    """

    def _check(field, value, error):
        # Validate flows
        for flow in value:
            if "origin" in flow and flow["origin"] not in compartment_names:
                error(field, "From compartment name not found in compartment types")
            if "to" in flow and flow["to"] not in compartment_names:
                error(field, "To compartment name not found in compartment types")

    return _check


def check_initial_conditions(starting_population, compartment_names):
    """
    Ensure initial conditions are well formed, and do not exceed population numbers.
    """

    def _check(field, value, error):
        try:
            is_pop_too_small = sum(value.values()) > starting_population
            if is_pop_too_small:
                error(
                    field,
                    "Initial condition population exceeds total starting population.",
                )

            if not all([c in compartment_names for c in value.keys()]):
                error(
                    field,
                    "Initial condition compartment name is not one of the listed compartment types",
                )
        except TypeError:
            error(field, "Could not check initial conditions.")

    return _check


def check_times(field, value, error):
    """
    Ensure times are sorted in ascending order.
    """
    if sorted(value) != value:
        error(field, "Integration times are not in order")


class ValidationException(Exception):
    """
    Raised when user-defined data is found to be invalid.
    """

    pass



def replicate_compartment(
    n_replications,
    current_compartments,
    compartment_stem,
    infectious_compartments,
    initial_populations,
    infectious=False,
    infectious_seed=0.0,
):
    """
    Implements n repeated sequential compartments of a certain type
    Also returns the names of the infectious compartments and the initial population evenly distributed across the
    replicated infectious compartments.

    :params:
        n_replications: int
            number of times compartment is to be repeated
        current_compartments: list
            all the compartments implemented up to this point
        compartment_stem: str
            name of compartment that needs to be replicated
        infectious_compartments: list
            all the infectious compartments implemented up to this point
        initial_population: float

        infectious: bool
            whether the compartment currently being replicated is infectious
        infectious_seed: float
            number of persons to contribute to the infectious seed
    :return:
        updated list of model compartments
        updated list of infectious compartments
        updated dictionary of initial conditions
    """

    # Add the compartment names to the working list of compartments
    compartments_to_add = (
        [compartment_stem]
        if n_replications == 1
        else [compartment_stem + "_" + str(i_comp + 1) for i_comp in range(n_replications)]
    )

    # Add the compartment names to the working list of infectious compartments, if the compartment is infectious
    infectious_compartments_to_add = compartments_to_add if infectious else []

    # Add the infectious population to the initial conditions
    if infectious_seed == 0.0:
        init_pop = {}
    elif n_replications == 1:
        init_pop = {compartment_stem: infectious_seed}
    else:
        init_pop = {
            compartment_stem + "_" + str(i_infectious + 1): infectious_seed / float(n_replications)
            for i_infectious in range(n_replications)
        }
    initial_populations.update(init_pop)

    return (
        current_compartments + compartments_to_add,
        infectious_compartments + infectious_compartments_to_add,
        initial_populations,
    )



"""
Data manipulation functions
"""


def increment_list_by_index(list_to_increment, index_to_increment, increment_value):
    """
    very simple but general method to increment the odes by a specified value

    :param list_to_increment: list
        the list to be incremented, expected to be the list of ODEs
    :param index_to_increment: int
        the index of the list that needs to be incremented
    :param increment_value: float
        the value to increment the list by
    :return: list
        the list after incrementing
    """
    list_to_increment[index_to_increment] += increment_value
    return list_to_increment


def normalise_dict(value_dict):
    """
    normalise the values from a list using the total of all values, i.e. returning dictionary whose keys sum to one with
        same ratios between all the values in the dictionary after the function has been applied

    :param value_dict: dict
        dictionary whose values will be adjusted
    :return: dict
        same dictionary after values have been normalised to the total of the original values
    """
    return {key: value_dict[key] / sum(value_dict.values()) for key in value_dict}


def order_dict_by_keys(input_dict):
    """
    sort the input dictionary keys and return two separate lists with keys and values as lists with corresponding
        elements

    :param input_dict: dict
        dictionary to be sorted
    :return:
        :dict_keys: list
            sorted list of what were the dictionary keys
        : list
            values applicable to the sorted list of dictionary keys
    """
    dict_keys = list(input_dict.keys())
    dict_keys.sort()
    return dict_keys, [input_dict[key] for key in dict_keys]


def element_list_multiplication(list_1, list_2):
    """
    multiply elements of two lists to return another list with the same dimensions

    :param list_1: list
        first list of numeric values to be multiplied
    :param list_2: list
        second list of numeric values to be multiplied
    :return: list
        resulting list populated with the multiplied values
    """
    return [a * b for a, b in zip(list_1, list_2)]


def element_list_division(list_1, list_2):
    """
    divide elements of two lists to return another list with the same dimensions

    :param list_1: list
        first list of numeric values for numerators of division
    :param list_2: list
        second list of numeric values for denominators of division
    :return: list
        list populated with quotient values
    """
    return [a / b for a, b in zip(list_1, list_2)]


def convert_boolean_list_to_indices(list_of_booleans):
    """
    take a list of boolean values and return the indices of the elements containing True

    :param list_of_booleans: list
        sequence of booleans
    :return: list with integer values
        list of the values that were True in the input list_of_booleans
    """
    return [n_element for n_element, element in enumerate(list_of_booleans) if element]


def create_cumulative_dict(dict_of_props):

    cumulative_dict_of_props = {}
    cumulative_prop = 0.0
    for stratum in dict_of_props:
        cumulative_prop += dict_of_props[stratum]
        cumulative_dict_of_props[stratum] = cumulative_prop
    return cumulative_dict_of_props


def find_first_list_element_above(a_list, value):
    """
    Simple method to return the index of the first element of a list that is greater than a specified value.

    Args:
        a_list: List of floats
        value: The value that the element must be greater than
    """
    if max(a_list) <= value:
        ValueError("The requested value is greater than max(a_list)")

    for i, val in enumerate(a_list):
        if val > value:
            return i


def remove_multiple_elements_from_list(a_list, indices_to_be_removed):
    """
    remove list elements according to a list of indices to be removed from that list

    :param a_list: list
        list to be processed
    :param indices_to_be_removed: list
        list of the elements that are no longer needed
    """
    return [a_list[i] for i in range(len(a_list)) if i not in indices_to_be_removed]



import copy
import itertools
from typing import List, Dict, Callable, Optional, Union

import numpy as np

from summer.legacy.constants import (
    Flow,
    BirthApproach,
    IntegrationType,
)
from summer.legacy.model.epi_model import EpiModel
from summer.legacy.model.utils.validation import validate_stratify
from summer.legacy.compartment import Compartment
from summer.legacy.flow import AgeingFlow
from summer.legacy.stratification import (
    Stratification,
    get_stratified_compartment_names,
    get_stratified_compartment_values,
)


class StratifiedModel(EpiModel):
    """
    Stratified compartmental model.
    """

    DEFAULT_DISEASE_STRAIN = "default"
    DEFAULT_MIXING_MATRIX = np.array([[1]])

    def __init__(
        self,
        times: List[float],
        compartment_names: List[str],
        initial_conditions: Dict[str, float],
        parameters: Dict[str, float],
        requested_flows: List[dict],
        infectious_compartments: List[str],
        birth_approach: str,
        entry_compartment: str,
        starting_population: int,
    ):
        super().__init__(
            times=times,
            compartment_names=compartment_names,
            initial_conditions=initial_conditions,
            parameters=parameters,
            requested_flows=requested_flows,
            infectious_compartments=infectious_compartments,
            birth_approach=birth_approach,
            entry_compartment=entry_compartment,
            starting_population=starting_population,
        )
        # Keeps track of Stratifications that have been applied.
        self.stratifications = []

        # Keeps track of original, pre-stratified compartment names.
        self.original_compartment_names = [Compartment.deserialize(n) for n in compartment_names]

        # A list of the different sub-categories ('strains') of the diease that we are modelling.
        self.disease_strains = [self.DEFAULT_DISEASE_STRAIN]

        # Strata-based multipliers for compartmental infectiousness levels.
        self.infectiousness_levels = {}

        # Compartment-based overwrite values for compartmental infectiousness levels.
        self.individual_infectiousness_adjustments = []

        # Mixing matrices a list of NxN arrays used to calculate force of infection.
        self.mixing_matrices = []

        # A list of dicts that has the strata required to match a row in the mixing matrix.
        self.mixing_categories = [{}]

    def stratify(
        self,
        stratification_name: str,
        strata_request: List[str],
        compartments_to_stratify: List[str],
        comp_split_props: Dict[str, float] = {},
        flow_adjustments: Dict[str, Dict[str, float]] = {},
        infectiousness_adjustments: Dict[str, float] = {},
        mixing_matrix: Union[np.ndarray, Callable[[float], np.ndarray]] = None,
    ):
        """
        Apply a stratification to the model's compartments.

        stratification_name: The name of the stratification
        strata_names: The names of the strata to apply
        compartments_to_stratify: The compartments that will have the stratification applied. Falsey args interpreted as "all".
        comp_split_props: Request to split existing population in the compartments according to specific proportions
        """
        validate_stratify(
            self,
            stratification_name,
            strata_request,
            compartments_to_stratify,
            comp_split_props,
            flow_adjustments,
            infectiousness_adjustments,
            mixing_matrix,
        )
        strat = Stratification(
            name=stratification_name,
            strata=strata_request,
            compartments=compartments_to_stratify,
            comp_split_props=comp_split_props,
            flow_adjustments=flow_adjustments,
        )
        self.stratifications.append(strat)

        # Add this stratification's mixing matrix if a new one is provided.
        if mixing_matrix is not None:
            assert not strat.is_strain(), "Strains cannot have a mixing matrix."
            # Only allow mixing matrix to be supplied if there is a complete stratification.
            assert (
                compartments_to_stratify == self.original_compartment_names
            ), "Mixing matrices only allowed for full stratification."
            self.mixing_matrices.append(mixing_matrix)
            self.mixing_categories = strat.update_mixing_categories(self.mixing_categories)

        # Prepare infectiousness levels for force of infection adjustments.
        self.infectiousness_levels[strat.name] = infectiousness_adjustments

        if strat.is_strain():
            # Track disease strain names, overriding default values.
            self.disease_strains = strat.strata

        # Stratify compartments, split according to split_proportions
        prev_compartment_names = copy.copy(self.compartment_names)
        self.compartment_names = get_stratified_compartment_names(strat, self.compartment_names)
        self.compartment_values = get_stratified_compartment_values(
            strat, prev_compartment_names, self.compartment_values
        )
        for idx, c in enumerate(self.compartment_names):
            c.idx = idx

        # Stratify flows
        prev_flows = self.flows
        self.flows = []
        for flow in prev_flows:
            self.flows += flow.stratify(strat)

        if strat.is_ageing():
            ageing_flows, ageing_params = AgeingFlow.create(
                strat, prev_compartment_names, self.get_parameter_value
            )
            self.flows += ageing_flows
            self.parameters.update(ageing_params)

        # Update indicies used by flows to lookup compartment values.
        compartment_idx_lookup = {name: idx for idx, name in enumerate(self.compartment_names)}
        for flow in self.flows:
            flow.update_compartment_indices(compartment_idx_lookup)

    def add_extra_flow(
        self,
        flow: dict,
        source_strata: dict,
        dest_strata: dict,
        expected_flow_count: int = None,
    ):
        """
        Add a new flow from a set of source compartments to a set of dest compartments.
        This can be done before or after stratification(s).

        The supplied strata will be used to find the source and destination compartments.
        There must be an equal number of source and destination compartments.
        Any unspecified strata will have flows created in parallel.

        flow: The flow to add, same format as when construction the model.
        source_strata: Source strata used to find source compartments.
        dest_strata: Destination strata used to find destination compartments.
        expected_count: (Optional) Expected number of flows to be created.
        """
        assert flow["type"] in Flow.TRANSITION_FLOWS, "Can only add extra transition flows."
        flows_to_add = []
        source_comps = [
            comp for comp in self.compartment_names if comp.is_match(flow["origin"], source_strata)
        ]
        dest_comps = [
            comp for comp in self.compartment_names if comp.is_match(flow["to"], dest_strata)
        ]
        count_dest_comps = len(dest_comps)
        count_source_comps = len(source_comps)
        msg = f"Expected equal number of source and dest compartmentss, but got {count_source_comps} source and {count_dest_comps} dest."
        assert count_dest_comps == count_source_comps, msg
        for source_comp, dest_comp in zip(source_comps, dest_comps):
            flow_to_add = {
                **flow,
                "origin": source_comp.serialize(),
                "to": dest_comp.serialize(),
            }
            flows_to_add.append(flow_to_add)

        if expected_flow_count is not None:
            # Check that we added the expected number of flows.
            actual_flow_count = len(flows_to_add)
            msg = f"Expected to add {expected_flow_count} flows but added {actual_flow_count}"
            assert actual_flow_count == expected_flow_count, msg

        # Add the new flows to the model
        for flow_to_add in flows_to_add:
            self._add_flow(flow_to_add)

        # Update flow compartment idxs for faster lookups.
        self._update_flow_compartment_indices()

    def get_mixing_matrix(self, time: float) -> np.ndarray:
        """
        Returns the final mixing matrix for a given time.
        """
        mixing_matrix = self.DEFAULT_MIXING_MATRIX
        for mm_func in self.mixing_matrices:
            # Assume each mixing matrix is either an np.ndarray or a function of time that returns one.
            mm = mm_func(time) if callable(mm_func) else mm_func
            # Get Kronecker product of old and new mixing matrices.
            mixing_matrix = np.kron(mixing_matrix, mm)

        return mixing_matrix

    def prepare_force_of_infection(self):
        """
        Pre-run calculations to help determine force of infection multiplier at runtime.

        We start with a set of "mixing categories". These categories describe groups of compartments.
        For example, we might have the stratifications age {child, adult} and location {work, home}.
        In this case, the mixing categories would be {child x home, child x work, adult x home, adult x work}.
        Mixing categories are only created when a mixing matrix is supplied during stratification.

        There is a mapping from every compartment to a mixing category.
        This is only true if mixing matrices are supplied only for complete stratifications.
        There are `num_cats` categories and `num_comps` compartments.
        The category matrix is a (num_cats x num_comps) matrix of 0s and 1s, with a 1 when the compartment is in a given category.
        We expect only one category per compartment, but there may be many compartments per category.

        We can multiply the category matrix by the vector of compartment sizes to get the total number of people
        in each mixing category.

        We also create a vector of values in [0, inf) which describes how infectious each compartment is: compartment_infectiousness
        We can use this vector plus the compartment sizes to get the 'effective' number of infectious people per compartment.

        We can use the 'effective infectious' compartment sizes, plus the mixing category matrix
        to get the infected population per mixing category.

        Now that we know:
            - the total population per category
            - the infected population per category
            - the inter-category mixing coefficients (mixing matrix)

        We can calculate the infection density or frequency per category.
        Finally, at runtime, we can lookup which category a given compartment is in and look up its infectious multiplier (density or frequency).
        """
        # Find out the relative infectiousness of each compartment, for each strain.
        # If no strains have been create, we assume a default strain name.
        self.compartment_infectiousness = {
            strain_name: self.get_compartment_infectiousness(strain_name)
            for strain_name in self.disease_strains
        }

        # Create a matrix that tracks which categories each compartment is in.
        # A matrix with size (num_cats x num_comps).
        num_comps = len(self.compartment_names)
        num_categories = len(self.mixing_categories)
        self.category_lookup = {}  # Map compartments to categories.
        self.category_matrix = np.zeros((num_categories, num_comps))
        for i, category in enumerate(self.mixing_categories):
            for j, comp in enumerate(self.compartment_names):
                if all(comp.has_stratum(k, v) for k, v in category.items()):
                    self.category_matrix[i][j] = 1
                    self.category_lookup[j] = i

    def get_compartment_infectiousness(self, strain: str):
        """
        Returns a vector of floats, each representing the relative infectiousness of each compartment.
        If a strain name is provided, find the infectiousness factor *only for that strain*.
        """
        # Figure out which compartments should be infectious
        infectious_mask = np.array(
            [c.has_name_in_list(self.infectious_compartments) for c in self.compartment_names]
        )
        # Find the infectiousness multipliers for each compartment being implemented in the model.
        # Start from assumption that each compartment is not infectious.
        compartment_infectiousness = np.zeros(self.compartment_values.shape)
        # Set all infectious compartments to be equally infectious.
        compartment_infectiousness[infectious_mask] = 1

        # Multiply by used-requested factors for particular strata.
        for idx, compartment in enumerate(self.compartment_names):
            for strat_name, strat_multipliers in self.infectiousness_levels.items():
                for stratum_name, multiplier in strat_multipliers.items():
                    if compartment.has_stratum(strat_name, stratum_name):
                        compartment_infectiousness[idx] *= multiplier

        # Override with user-requested values for particular compartments.
        # FIXME: This is yucky and requires user to hack on model post/pre stratification.
        for adj in self.individual_infectiousness_adjustments:
            comp_name = adj["comp_name"]
            comp_strata = adj["comp_strata"]
            value = adj["value"]
            for idx, comp in enumerate(self.compartment_names):
                if not comp.has_name(comp_name):
                    continue

                if not all(comp.has_stratum(k, v) for k, v in comp_strata.items()):
                    continue

                compartment_infectiousness[idx] = value

        if strain != self.DEFAULT_DISEASE_STRAIN:
            # Filter out all values that are not in the given strain.
            strain_mask = np.zeros(self.compartment_values.shape)
            for idx, compartment in enumerate(self.compartment_names):
                if compartment.has_stratum("strain", strain):
                    strain_mask[idx] = 1

            compartment_infectiousness *= strain_mask

        return compartment_infectiousness

    def get_force_idx(self, source: Compartment):
        """
        Returns the index of the source compartment in the infection multiplier vector.
        """
        return self.category_lookup[source.idx]

    def find_infectious_population(self, time: float, compartment_values: np.ndarray):
        """
        Finds the effective infectious population.
        """
        mixing_matrix = self.get_mixing_matrix(time)

        # Calculate total number of people per category.
        # A vector with size (num_cats x 1).
        comp_values_transposed = compartment_values.reshape((compartment_values.shape[0], 1))
        self.category_populations = np.matmul(self.category_matrix, comp_values_transposed)

        # Calculate infectious populations for each strain.
        # Infection density / frequency is the infectious multiplier for each mixing category, calculated for each strain.
        self.infection_density = {}
        self.infection_frequency = {}
        for strain in self.disease_strains:
            strain_compartment_infectiousness = self.compartment_infectiousness[strain]

            # Calculate total infected people per category, including adjustment factors.
            # Returns a vector with size (num_cats x 1).
            infected_values = compartment_values * strain_compartment_infectiousness
            infected_values_transposed = infected_values.reshape((infected_values.shape[0], 1))
            infectious_populations = np.matmul(self.category_matrix, infected_values_transposed)
            self.infection_density[strain] = np.matmul(mixing_matrix, infectious_populations)

            # Calculate total infected person frequency per category, including adjustment factors.
            # A vector with size (num_cats x 1).
            category_frequency = infectious_populations / self.category_populations
            self.infection_frequency[strain] = np.matmul(mixing_matrix, category_frequency)

    def get_infection_frequency_multiplier(self, source: Compartment, dest: Compartment):
        """
        Get force of infection multiplier for a given compartment,
        using the 'infection frequency' calculation.
        """
        idx = self.get_force_idx(source)
        strain = dest._strat_values.get("strain", self.DEFAULT_DISEASE_STRAIN)
        return self.infection_frequency[strain][idx][0]

    def get_infection_density_multiplier(self, source: Compartment, dest: Compartment):
        """
        Get force of infection multiplier for a given compartment,
        using the 'infection density' calculation.
        """
        idx = self.get_force_idx(source)
        strain = dest._strat_values.get("strain", self.DEFAULT_DISEASE_STRAIN)
        return self.infection_density[strain][idx][0]



"""
Constants used in building SUMMER models.
"""


class FlowAdjustment:
    """
    A type of adjustment to a stratified flow.
    """

    OVERWRITE = "OVERWRITE"  # Overwrite with a value
    MULTIPLY = "MULTIPLY"  # Multiply with a value
    COMPOSE = "COMPOSE"  # Compose with a function


class Compartment:
    """
    A model compartment.
    """

    SUSCEPTIBLE = "susceptible"
    EARLY_LATENT = "early_latent"
    LATE_LATENT = "late_latent"
    EXPOSED = "exposed"
    INFECTIOUS = "infectious"
    EARLY_INFECTIOUS = "infectious"
    RECOVERED = "recovered"
    LATE_INFECTIOUS = "late"
    ON_TREATMENT = "on_treatment"
    LTBI_TREATED = "ltbi_treated"
    DETECTED = "detected"


class Flow:
    """
    A type of flow between model compartments
    """

    # Transition flows: where people are moving from one compartment to another
    STANDARD = "standard"
    INFECTION_FREQUENCY = "infection_frequency"
    INFECTION_DENSITY = "infection_density"
    TRANSITION_FLOWS = (
        STANDARD,
        INFECTION_FREQUENCY,
        INFECTION_DENSITY,
    )
    INFECTION_FLOWS = (INFECTION_DENSITY, INFECTION_FREQUENCY)

    # Entry flows: where people are entering the population (births, imports)
    BIRTH = "birth"
    IMPORT = "import"
    ENTRY_FLOWS = (BIRTH, IMPORT)

    # Exit flows: where people are leaving the population (deaths, exports)
    DEATH = "compartment_death"
    UNIVERSAL_DEATH = "universal_death"
    DEATH_FLOWS = (DEATH, UNIVERSAL_DEATH)
    EXIT_FLOWS = (DEATH, UNIVERSAL_DEATH)


class BirthApproach:
    """
    Options for birth rate settings in model
    """

    NO_BIRTH = "no_birth"
    ADD_CRUDE = "add_crude_birth_rate"
    REPLACE_DEATHS = "replace_deaths"


class IntegrationType:
    """
    Options for ODE solver used by model
    """

    ODE_INT = "odeint"
    SOLVE_IVP = "solve_ivp"
    EULER = "euler"
    RUNGE_KUTTA = "rk4"



from typing import List, Tuple, Dict, Callable

from summer.legacy.stratification import Stratification
from summer.legacy.constants import FlowAdjustment

from .flow import BaseFlow


class BaseEntryFlow(BaseFlow):
    """
    A flow where people enter the destination compartment, but there is no source.
    Eg. births, importation.
    """

    def update_compartment_indices(self, mapping: Dict[str, float]):
        """
        Update index which maps flow compartments to compartment value array.
        """
        self.dest.idx = mapping[self.dest]

    def stratify(self, strat: Stratification) -> List[BaseFlow]:
        """
        Returns a list of new, stratified entry flows to replace the current flow.
        """
        if not self.dest.has_name_in_list(strat.compartments):
            # Flow destination is not stratified, do not stratify this flow.
            return [self]

        new_flows = []
        for stratum in strat.strata:
            adjustment = None
            if strat.is_ageing():
                # Use special rules for ageing.
                if stratum == "0":
                    # Babies get born at age 0, add a null op to prevent default behaviour.
                    adjustment = (FlowAdjustment.MULTIPLY, 1)
                else:
                    # Babies do not get born at any other age.
                    adjustment = (FlowAdjustment.OVERWRITE, 0)
            else:
                # Not an ageing stratification, check for user-specified flow adjustments.
                adjustment = strat.get_flow_adjustment(self.dest, stratum, self.param_name)

            if not adjustment:
                # Default to equally dividing entry population between all strata.
                num_strata = len(strat.strata)
                entry_fraction = 1.0 / num_strata
                adjustment = (FlowAdjustment.MULTIPLY, entry_fraction)

            new_adjustments = [*self.adjustments, adjustment]
            new_dest = self.dest.stratify(strat.name, stratum)
            new_flow = self.copy(
                dest=new_dest,
                param_name=self.param_name,
                param_func=self.param_func,
                adjustments=new_adjustments,
            )
            new_flows.append(new_flow)

        return new_flows



from typing import List, Tuple, Dict, Callable
from summer.legacy.constants import Flow as FlowType
from summer.legacy.compartment import Compartment
from summer.legacy.stratification import Stratification

from .base import BaseEntryFlow


class ReplacementBirthFlow(BaseEntryFlow):
    """
    Calculates births by replacing total deaths.
    """

    type = FlowType.BIRTH

    def __init__(self, dest: Compartment, get_total_deaths: Callable[[], float], adjustments=[]):

        assert type(dest) is Compartment
        self.adjustments = adjustments
        self.param_name = "total_deaths"
        self.dest = dest
        self.get_total_deaths = get_total_deaths

    def param_func(self, name, time):
        """
        Hack so that `get_weight_value` will use `get_total_deaths`
        and also apply any adjustments.
        """
        assert time is None, "Total deaths is not a function of time."
        if name == "total_deaths":
            return self.get_total_deaths()
        else:
            raise ValueError("Cannot use param_func for replace deaths.")

    def get_net_flow(self, compartment_values, time):
        return self.get_weight_value(None)

    def copy(self, **kwargs):
        """
        Creates a modified copy of the flow for stratification.
        """
        return ReplacementBirthFlow(
            dest=kwargs["dest"],
            param_name=kwargs["param_name"],
            param_func=kwargs["param_func"],
            adjustments=kwargs["adjustments"],
        )

    def __repr__(self):
        return f"<ReplacementBirthFlow to {self.dest} with {self.param_name}>"



from typing import List, Tuple, Dict, Callable

from summer.legacy.constants import FlowAdjustment
from summer.legacy.stratification import Stratification
from .flow import BaseFlow


class BaseTransitionFlow(BaseFlow):
    """
    A flow where people move from the source compartment, to the destination.
    Eg. infection, recovery, progress of disease.
    """

    def update_compartment_indices(self, mapping: Dict[str, float]):
        """
        Update index which maps flow compartments to compartment value array.
        """
        self.source.idx = mapping[self.source]
        self.dest.idx = mapping[self.dest]

    def stratify(self, strat: Stratification) -> List[BaseFlow]:
        """
        Returns a list of new, stratified flows to replace the current flow.
        """
        is_source_strat = self.source.has_name_in_list(strat.compartments)
        is_dest_strat = self.dest.has_name_in_list(strat.compartments)
        if not (is_dest_strat or is_source_strat):
            # Flow is not stratified, do not stratify this flow.
            return [self]

        new_flows = []
        for stratum in strat.strata:
            # Find new compartments
            if is_source_strat:
                new_source = self.source.stratify(strat.name, stratum)
            else:
                new_source = self.source

            if is_dest_strat:
                new_dest = self.dest.stratify(strat.name, stratum)
            else:
                new_dest = self.dest

            # Find flow adjustments to apply to the new stratified flows.
            # First, we try to find an adjustment for the source compartment.
            # This is for when the source has the required stratifications and the destination does not.
            # For example - people recovering from I -> R with multiple I strata, all with different recovery rates.
            adjustment = strat.get_flow_adjustment(self.source, stratum, self.param_name)
            if not adjustment:
                # Otherwise, try find an adjustment for the destination compartment.
                # This is for when the destination has the required stratifications and the source does not.
                # For example - people recovering from I -> R with multiple R strata, with different recovery proportions.
                adjustment = strat.get_flow_adjustment(self.dest, stratum, self.param_name)

            # Should we apply an adjustment to conserve the number of people?
            should_apply_conservation_split = (
                (not strat.is_strain())
                and (not adjustment)
                and (is_dest_strat and not is_source_strat)
            )
            if should_apply_conservation_split:
                # If the source is stratified but not the destination, then we need to account
                # for the resulting fan-out of flows by reducing the flow rate.
                # We don't do this for strains because this effect is already
                # captured by the infecitousness multiplier.
                num_strata = len(strat.strata)
                entry_fraction = 1.0 / num_strata
                adjustment = (FlowAdjustment.MULTIPLY, entry_fraction)

            if adjustment:
                new_adjustments = [*self.adjustments, adjustment]
            else:
                new_adjustments = self.adjustments

            new_flow = self.copy(
                source=new_source,
                dest=new_dest,
                param_name=self.param_name,
                param_func=self.param_func,
                adjustments=new_adjustments,
            )
            new_flows.append(new_flow)

        return new_flows



from abc import ABC, abstractmethod
from typing import List, Tuple, Dict, Callable

import numpy as np

from summer.legacy.stratification import Stratification
from summer.legacy.constants import FlowAdjustment


class BaseFlow(ABC):
    """
    Abstract base class for all flows.
    A flow represents the movement of people from one compartment to another.
    """

    param_name = None
    param_func = None
    adjustments = None

    @property
    @abstractmethod
    def type(self):
        """Returns the type of flow"""
        pass

    def get_weight_value(self, time: float):
        """
        Returns the flow's weight at a given time.
        Applies any stratification adjustments to the base parameter.
        """
        value = self.param_func(self.param_name, time)
        return self._apply_adjustments(value, time)

    def _apply_adjustments(self, value: float, time: float):
        for adjust_type, adjust_value in self.adjustments:
            if adjust_type == FlowAdjustment.COMPOSE:
                # Multiply by time variant function value.
                value *= self.param_func(adjust_value, time)
            elif adjust_type == FlowAdjustment.OVERWRITE:
                # Overwrite with adjustment value.
                value = adjust_value
            elif adjust_type == FlowAdjustment.MULTIPLY:
                # Multiply by adjustment value.
                value *= adjust_value

        return value

    @abstractmethod
    def update_compartment_indices(self, mapping: Dict[str, float]):
        """
        Update index which maps flow compartments to compartment value array.
        """
        pass

    @abstractmethod
    def get_net_flow(self, compartment_values: np.ndarray, time: float) -> float:
        """
        Returns the net flow value at a given time.
        """
        pass

    @abstractmethod
    def stratify(self, strat: Stratification) -> list:
        """
        Returns a list of new, stratified flows to replace the current flow.
        """
        pass

    @abstractmethod
    def copy(self, **kwargs):
        """
        Creates a modified copy of the flow for stratification.
        """
        pass

    @abstractmethod
    def __repr__(self):
        """
        Returns a text representation of the flow.
        """
        pass



from typing import List, Tuple, Dict, Callable

import numpy as np
from numba import jit

from summer.legacy.constants import Flow as FlowType
from summer.legacy.compartment import Compartment
from summer.legacy.stratification import Stratification

from .base import BaseEntryFlow


class CrudeBirthFlow(BaseEntryFlow):
    """
    Calculates births using a 'crude birth rate' method.
    """

    type = FlowType.BIRTH

    def __init__(
        self,
        dest: Compartment,
        param_name: str,
        param_func: Callable[[str, float], float],
        adjustments: list = [],
    ):
        assert type(dest) is Compartment
        self.adjustments = adjustments
        self.dest = dest
        self.param_name = param_name
        self.param_func = param_func

    def get_net_flow(self, compartment_values, time):
        parameter_value = self.get_weight_value(time)
        population = _find_sum(compartment_values)
        return parameter_value * population

    def copy(self, **kwargs):
        """
        Creates a modified copy of the flow for stratification.
        """
        return CrudeBirthFlow(
            dest=kwargs["dest"],
            param_name=kwargs["param_name"],
            param_func=kwargs["param_func"],
            adjustments=kwargs["adjustments"],
        )

    def __repr__(self):
        return f"<CrudeBirthFlow to {self.dest} with {self.param_name}>"


@jit(nopython=True)
def _find_sum(compartment_values: np.ndarray) -> float:
    return compartment_values.sum()



from typing import List, Dict


class Compartment:
    """
    A compartment in the compartmental model.
    Each compartment is full of people and can be stratified into many smaller compartments.
    """

    def __init__(
        self,
        name: str,
        strat_names: List[str] = [],
        strat_values: Dict[str, str] = {},
    ):
        assert set(strat_names) == set(strat_values.keys()), "Mismatch in Compartment strat values."
        assert type(name) is str, "Name must be a string, not %s." % type(name)
        # Internal values, must not be modified by external code.
        self._name = name
        self._strat_names = tuple(strat_names)
        self._strat_values = strat_values
        self._str = self.serialize()
        self.idx = None

    def is_match(self, name: str, strata: dict) -> bool:
        """
        Returns True if this compartment matches the supplied name and strata.
        A partial strata match returns True.
        """
        is_name_match = name == self._name
        is_strata_match = all([self.has_stratum(k, v) for k, v in strata.items()])
        return is_name_match and is_strata_match

    def has_name(self, comp):
        """
        Returns True if this compartment has the same root name as another.
        """
        if type(comp) is str:
            return self._name == comp
        else:
            return self._name == comp._name

    def has_name_in_list(self, comps: list):
        """
        Returns True if this compartment has the same root name
        as any compartment in the provided list.
        """
        return any(self.has_name(c) for c in comps)

    def has_stratum(self, name: str, value: str):
        return self._strat_values.get(name) == value

    def get_strata(self) -> List[str]:
        return [self._serialize_stratum(s) for s in self._strat_names]

    def serialize(self) -> str:
        """Return string representation of the compartment"""
        return "X".join([self._name, *self.get_strata()])

    def stratify(self, stratify_name: str, stratum_name: str):
        new_strat_names = [*self._strat_names, stratify_name]
        new_strat_values = {**self._strat_values, stratify_name: stratum_name}
        return Compartment(
            name=self._name,
            strat_names=new_strat_names,
            strat_values=new_strat_values,
        )

    @staticmethod
    def deserialize(s: str):
        """Convert string into Compartment"""
        parts = s.split("X")
        name = parts[0]
        strat_names = []
        strat_values = {}
        for strat in parts[1:]:
            s_name, s_value = Compartment._deserialize_stratum(strat)
            strat_names.append(s_name)
            strat_values[s_name] = s_value

        return Compartment(name, strat_names, strat_values)

    @staticmethod
    def _deserialize_stratum(stratum_string):
        stratum_parts = stratum_string.split("_")
        name = stratum_parts[0]
        value = "_".join(stratum_parts[1:])
        return name, value

    def _serialize_stratum(self, stratum_name):
        return f"{stratum_name}_{self._strat_values[stratum_name]}"

    def __str__(self) -> str:
        return self._str

    def __repr__(self) -> str:
        return self._str

    def __eq__(self, obj):
        """
        Equality operator override
        """
        is_same_as_str = type(obj) is str and obj == self._str
        is_same_as_comp = type(obj) is Compartment and obj._str == self._str
        return is_same_as_str or is_same_as_comp

    def __hash__(self):
        """
        Hash operator override
        """
        return hash(self._str)



"""
This module presents classes which are used to define stratifications, which can be applied to the model.
"""
from typing import List, Dict, Callable, Union, Optional

import numpy as np

from summer.compartment import Compartment
from summer.adjust import Multiply, Overwrite


Adjustment = Union[Multiply, Overwrite]
MixingMatrix = Union[np.ndarray, Callable[[float], np.ndarray]]


class Stratification:
    """
    A generic stratification applied to the compartmental model.

    Args:
        name: The name of the stratification.
        strata: The new strata that will be created.
        compartments: The compartments which will be stratified.

    """

    _is_ageing = False
    _is_strain = False

    def __init__(
        self,
        name: str,
        strata: List[str],
        compartments: List[str],
    ):
        self.name = name
        self.strata = list(map(str, strata))
        self.compartments = [Compartment(c) if type(c) is str else c for c in compartments]

        # Split population evently by default.
        num_strata = len(self.strata)
        self.population_split = {s: 1 / num_strata for s in self.strata}

        # Flows are not adjusted by default.
        self.flow_adjustments = {}
        self.infectiousness_adjustments = {}

        # No heterogeneous mixing matrix by default.
        self.mixing_matrix = None

    def is_ageing(self) -> bool:
        """Returns ``True`` if this stratification represets a set of age groups with ageing dynamics"""
        return self._is_ageing

    def is_strain(self) -> bool:
        """Returns ``True`` if this stratification represets a set of disease strains"""
        return self._is_strain

    def set_population_split(self, proportions: Dict[str, float]):
        """
        Sets how the stratification will split the population between the strata.

        Args:
            proportions: A map of proportions to be assigned to each strata.

        The supplied proportions must:

            - Specify all strata
            - All be positive
            - Sum to 1 +/- 0.01

        """
        msg = f"All strata must be specified when setting population split: {proportions}"
        assert list(proportions.keys()) == self.strata, msg
        msg = f"All proportions must be >= 0 when setting population split: {proportions}"
        assert all([v >= 0 for v in proportions.values()]), msg
        msg = f"All proportions sum to 1+/-0.01 when setting population split: {proportions}"
        assert abs(1 - sum(proportions.values())) < 1e2, msg
        self.population_split = proportions

    def add_flow_adjustments(
        self,
        flow_name: str,
        adjustments: Dict[str, Adjustment],
        source_strata: Optional[Dict[str, str]] = None,
        dest_strata: Optional[Dict[str, str]] = None,
    ):
        """
        Add an adjustment of a flow to the stratification.
        You can use time-varying functions for infectiousness adjustments.

        It is possible to specify multiple conflicting flow adjustments for the same flow.
        In this case, only the last-created applicable adjustment will be chosen.

        Args:
            flow_name: The name of the flow to adjust.
            adjustments: An dict of adjustments to apply to the flow.
            source_strata (optional): A whitelist of strata to filter the target flow's source compartments.
            dest_strata (optional): A whitelist of strata to filter the target flow's destination compartments.

        Example:
            Create an adjustment for the 'recovery' flow based on location::

                strat = Stratification(
                    name="location",
                    strata=["urban", "rural", "alpine"],
                    compartments=["S", "I", "R"]
                )
                strat.add_flow_adjustments("recovery", {
                    "urban": Multiply(1.5),
                    "rural": Multiply(0.8),
                    "alpine": None, # No adjustment
                })


        """
        source_strata = source_strata or {}
        dest_strata = dest_strata or {}
        msg = "You must specify all strata when adding flow adjustments."
        assert set(adjustments.keys()) == set(self.strata), msg

        assert all(
            [type(a) is Overwrite or type(a) is Multiply or a is None for a in adjustments.values()]
        ), "All flow adjustments must be Multiply, Overwrite or None."

        if flow_name not in self.flow_adjustments:
            self.flow_adjustments[flow_name] = []

        self.flow_adjustments[flow_name].append((adjustments, source_strata, dest_strata))

    def get_flow_adjustment(self, flow) -> dict:
        """
        Returns the most recently added flow adjustment that matches a given flow.
        """
        flow_adjustments = self.flow_adjustments.get(flow.name, [])
        matching_adjustment = None
        for adjustment, source_strata, dest_strata in flow_adjustments:
            # Skip any adjustments that don't match the flow's source compartment strata.
            msg = f"Source strata specified in flow adjustment of {self.name} but {flow.name} does not have a source"
            assert not (source_strata and not flow.source), msg
            if source_strata and flow.source and not flow.source.has_strata(source_strata):
                continue

            # Skip any adjustments that don't match the flow's dest compartment strata.
            msg = f"Dest strata specified in flow adjustment of {self.name} but {flow.name} does not have a dest"
            assert not (dest_strata and not flow.dest), msg
            if dest_strata and flow.dest and not flow.dest.has_strata(dest_strata):
                continue

            matching_adjustment = adjustment

        return matching_adjustment

    def add_infectiousness_adjustments(
        self, compartment_name: str, adjustments: Dict[str, Adjustment]
    ):
        """
        Add an adjustment of a compartment's infectiousness to the stratification.
        You cannot use time-varying functions for infectiousness adjustments.

        Args:
            compartment_name: The name of the compartment to adjust.
            adjustments: An dict of adjustments to apply to the compartment.

        Example:
            Create an adjustment for the 'I' compartment based on location::

                strat = Stratification(
                    name="location",
                    strata=["urban", "rural", "alpine"],
                    compartments=["S", "I", "R"]
                )
                strat.add_infectiousness_adjustments("I", {
                    "urban": adjust.Multiply(1.5),
                    "rural": adjust.Multiply(0.8),
                    "alpine": None, # No adjustment
                })

        """
        msg = "You must specify all strata when adding infectiousness adjustments."
        assert set(adjustments.keys()) == set(self.strata), msg

        assert all(
            [type(a) is Overwrite or type(a) is Multiply or a is None for a in adjustments.values()]
        ), "All infectiousness adjustments must be Multiply, Overwrite or None."

        msg = f"An infectiousness adjustment for {compartment_name} already exists for strat {self.name}"
        assert compartment_name not in self.infectiousness_adjustments, msg

        msg = "Cannot use time varying functions for infectiousness adjustments."
        assert not any([callable(a.param) for a in adjustments.values() if a])

        self.infectiousness_adjustments[compartment_name] = adjustments

    def set_mixing_matrix(self, mixing_matrix: MixingMatrix):
        """"""
        msg = "Strain stratifications cannot have a mixing matrix."
        assert not self.is_strain(), msg

        mm = mixing_matrix(0) if callable(mixing_matrix) else mixing_matrix

        msg = "Mixing matrix must be a NumPy array, or return a NumPy array."
        assert type(mm) is np.ndarray, msg

        num_strata = len(self.strata)
        msg = f"Mixing matrix must have shape ({num_strata}, {num_strata})"
        assert mm.shape == (num_strata, num_strata), msg

        self.mixing_matrix = mixing_matrix

    def _stratify_compartments(self, comps: List[Compartment]) -> List[Compartment]:
        """
        Stratify the model compartments into sub-compartments, based on the strata names,
        Only compartments specified in the stratification's definition will be stratified.
        Returns the new compartments.
        """
        new_comps = []
        for old_comp in comps:
            should_stratify = old_comp.has_name_in_list(self.compartments)
            if should_stratify:
                for stratum in self.strata:
                    new_comp = old_comp.stratify(self.name, stratum)
                    new_comps.append(new_comp)
            else:
                new_comps.append(old_comp)

        return new_comps

    def _stratify_compartment_values(
        self, comps: List[Compartment], comp_values: np.ndarray
    ) -> np.ndarray:
        """
        Stratify the model compartments into sub-compartments, based on the strata names provided,
        Split the population according to the provided proportions.
        Only compartments specified in the stratification's definition will be stratified.
        Returns the new compartment values.
        """
        assert len(comps) == len(comp_values)
        new_comp_values = []
        for idx in range(len(comp_values)):
            should_stratify = comps[idx].has_name_in_list(self.compartments)
            if should_stratify:
                for stratum in self.strata:
                    new_value = comp_values[idx] * self.population_split[stratum]
                    new_comp_values.append(new_value)
            else:
                new_comp_values.append(comp_values[idx])

        return np.array(new_comp_values)


class AgeStratification(Stratification):
    """
    A stratification that represets a set of age groups with ageing dynamics.

    Args:
        name: The name of the stratification.
        strata: The new strata that will be created.
        compartments: The compartments which will be stratified.


    Strata must be a list of strings or integers that represent all ages. For example, the strata
    [0, 10, 20, 30] will be interpreted as the age groups 0-9, 10-19, 20-29, 30+ respectively.

    Using this stratification will automatically add a set of ageing flows to the model,
    where the exit rate is proportional to the age group's time span. For example, in the 0-9 strata,
    there will be a flow created where 10% of occupants "age up" into the 10-19 strata each year.
    **Critically**, this feature assumes that each of the model's time steps represents a year.
    """

    _is_ageing = True

    def __init__(
        self,
        name: str,
        strata: List[str],
        compartments: List[str],
    ):
        try:
            _strata = sorted(map(int, strata))
        except:
            raise AssertionError("Strata much be a int-compatible")

        assert _strata[0] == 0, "First age strata must be 0"

        _strata = map(str, _strata)
        super().__init__(name, _strata, compartments)


class StrainStratification(Stratification):
    """
    A stratification that represets a set of disease strains.

    Args:
        name: The name of the stratification.
        strata: The new strata that will be created.
        compartments: The compartments which will be stratified.


    Each supplied strata will be interpreted as a different strain of the disease being modelled.
    This will mean that the force of infection calculations will consider each strain separately.

    Strain stratifications cannot use a mixing matrix
    """

    _is_strain = True



"""
Functions of functions for use in stratified models
"""


def create_multiplicative_function(multiplier):
    """
    return multiplication by a fixed value as a function

    :param multiplier: float
        value that the returned function multiplies by
    :return: function
        function that can multiply by the multiplier parameter when called
    """
    return lambda input_value, time: multiplier * input_value


def create_time_variant_multiplicative_function(time_variant_function):
    """
    similar to create_multiplicative_function, except that the value to multiply by can be a function of time, rather
        than a single value

    :param time_variant_function: function
        a function with the independent variable of time that returns the value that the input should be multiplied by
    :return: function
        function that will multiply the input value by the output value of the time_variant_function
    """
    return lambda input_value, time: time_variant_function(time) * input_value


def create_additive_function(increment):
    """
    return the addition of a fixed value as a function

    :param increment: float
        value that the returned function increments by
    :return: function
        function that can increment by the value parameter when called
    """
    return lambda value: value + increment


def create_sloping_step_function(start_x, start_y, end_x, end_y):
    """
    create sloping step function, returning start y-value for input values below starting x-value, ending y-value
        for input values above ending x-value and connecting slope through the middle

    :param start_x: float
        starting x-value
    :param start_y: float
        starting y-value
    :param end_x: float
        ending x-value
    :param end_y: float
        ending y-value
    :return: function
        sloping function as described above
    """
    gradient = (end_y - start_y) / (end_x - start_x)

    def step_function(age):
        if age < start_x:
            return start_y
        elif start_x <= age < end_x:
            return gradient * age + start_y - gradient * start_x
        elif end_x <= age:
            return end_y

    return step_function


def create_function_of_function(outer_function, inner_function):
    """
    function that can itself return a function that sequentially apply two functions and so can be used recursively to
        create a series of functions

    :param outer_function: function
        last function to be called
    :param inner_function: function
        first function to be called
    :return: function
        composite function that applies the inner and then the outer function, allowing the time parameter to be passed
            through if necessary
    """
    return lambda time: outer_function(inner_function(time), time)



from typing import List, Tuple, Dict, Callable

from summer.legacy.stratification import Stratification
from .flow import BaseFlow


class BaseExitFlow(BaseFlow):
    """
    A flow where people exit the source compartment, but there is no destination
    Eg. deaths, exportation
    """

    def update_compartment_indices(self, mapping: Dict[str, float]):
        """
        Update index which maps flow compartments to compartment value array.
        """
        self.source.idx = mapping[self.source]

    def stratify(self, strat: Stratification) -> List[BaseFlow]:
        """
        Returns a list of new, stratified exit flows to replace the current flow.
        """
        if not self.source.has_name_in_list(strat.compartments):
            # Flow source is not stratified, do not stratify this flow.
            return [self]

        new_flows = []
        for stratum in strat.strata:
            adjustment = strat.get_flow_adjustment(self.source, stratum, self.param_name)
            if adjustment:
                new_adjustments = [*self.adjustments, adjustment]
            else:
                new_adjustments = self.adjustments

            new_source = self.source.stratify(strat.name, stratum)
            new_flow = self.copy(
                source=new_source,
                param_name=self.param_name,
                param_func=self.param_func,
                adjustments=new_adjustments,
            )
            new_flows.append(new_flow)

        return new_flows



from typing import List, Tuple, Dict, Callable

from summer.legacy.constants import Flow as FlowType
from summer.legacy.compartment import Compartment
from summer.legacy.stratification import Stratification

from .standard import StandardFlow


class AgeingFlow(StandardFlow):
    def __repr__(self):
        return f"<AgeingFlow from {self.source} to {self.dest} with {self.param_name}>"

    @staticmethod
    def create(
        strat: Stratification,
        compartments: List[Compartment],
        param_func: Callable[[str, float], float],
    ):
        """
        Create inter-compartmental flows for ageing from one stratum to the next.
        The ageing rate is proportional to the width of the age bracket.
        It's assumed that both ages and model timesteps are in years.
        """
        assert strat.is_ageing()
        ageing_flows = []
        ageing_params = {}
        ages = list(sorted(map(int, strat.strata)))
        for age_idx in range(len(ages) - 1):
            start_age = int(ages[age_idx])
            end_age = int(ages[age_idx + 1])
            param_name = f"ageing{start_age}to{end_age}"
            ageing_rate = 1.0 / (end_age - start_age)
            ageing_params[param_name] = ageing_rate
            for comp in compartments:
                if not comp.has_name_in_list(strat.compartments):
                    # Don't include unstratified compartments
                    continue

                flow = AgeingFlow(
                    source=comp.stratify(strat.name, str(start_age)),
                    dest=comp.stratify(strat.name, str(end_age)),
                    param_name=param_name,
                    param_func=param_func,
                )
                ageing_flows.append(flow)

        return ageing_flows, ageing_params



from typing import List, Tuple, Dict

import numpy as np

from summer.legacy.constants import FlowAdjustment
from summer.legacy.compartment import Compartment

STRAIN_STRATA_NAME = "strain"
AGE_STRATA_NAME = "age"
OVERWRITE_CHARACTER = "W"


class Stratification:
    """
    A stratification applied to a the compartmental model.
    """

    def __init__(
        self,
        name: str,
        strata: List[str],
        compartments: List[str],
        comp_split_props: Dict[str, float],
        flow_adjustments: Dict[str, Dict[str, float]],
    ):
        self.name = name
        self.strata = list(map(str, strata))
        self.comp_split_props = get_all_proportions(self.strata, comp_split_props)
        self.flow_adjustments = parse_flow_adjustments(flow_adjustments)
        self.compartments = [Compartment(c) if type(c) is str else c for c in compartments]

    def is_ageing(self) -> bool:
        return self.name == AGE_STRATA_NAME

    def is_strain(self) -> bool:
        return self.name == STRAIN_STRATA_NAME

    def get_flow_adjustment(self, comp: Compartment, stratum: str, param_name: str):
        """
        Returns an adjustment tuple or None
        """
        param_adjs = self.flow_adjustments.get(param_name)
        if not param_adjs:
            # No adjustments for this parameter.
            return

        comp_adj = None
        for _comp_adj in param_adjs:
            if _comp_adj["strata"] == comp._strat_values:
                comp_adj = _comp_adj["adjustments"]

        if not comp_adj:
            # No adjustments for this compartment.
            return

        return comp_adj.get(stratum)

    def update_mixing_categories(
        self, old_mixing_categories: List[Dict[str, str]]
    ) -> List[Dict[str, str]]:
        """
        Returns a new list of mixing categories, with this stratification included.
        """
        new_mc = []
        for mc in old_mixing_categories:
            for stratum in self.strata:
                new_mc.append({**mc, self.name: stratum})

        return new_mc


def parse_flow_adjustments(flow_adjustments: Dict[str, Dict[str, float]]):
    """
    INPUT
    {
        "to_infectiousXagegroup_50": {
            <empty>: just use the non-adjusted version with the old name
            "non_symptW": 0.51, # Overwrite to get stratified verion
            "clinical": 0.51, # Multiply to get stratified verion
            "icu": "prop_sympt_non_hospital_50", # Apply a function to previous parameter to get stratified verion
        },
    }
    OUTPUT
    {
        "to_infectious": [
            {
                "strata": {"agegroup": "50"},
                "adjustments": {
                    "non_sympt": (FlowAdjustment.OVERWRITE, 0.51),
                    "clinical": (FlowAdjustment.MULTIPLY, 0.51),
                    "icu": (FlowAdjustment.COMPOSE, "prop_sympt_non_hospital_50"),
                }
            }
        ]
    }
    """
    parsed_flow_adjustments = {}
    for adjust_key, adjustments in flow_adjustments.items():
        parts = adjust_key.split("X")
        param_name = parts[0]
        if param_name not in parsed_flow_adjustments:
            parsed_flow_adjustments[param_name] = []

        parsed_strata = {}
        for strat in parts[1:]:
            strat_parts = strat.split("_")
            parsed_strata[strat_parts[0]] = "_".join(strat_parts[1:])

        parsed_adjustments = {}
        for k, v in adjustments.items():
            if type(v) is str:
                parsed_adjustments[k] = (FlowAdjustment.COMPOSE, v)
            elif k.endswith(OVERWRITE_CHARACTER):
                key = k[:-1]
                parsed_adjustments[key] = (FlowAdjustment.OVERWRITE, v)
            else:
                parsed_adjustments[k] = (FlowAdjustment.MULTIPLY, v)

        parsed_adjustment = {
            "strata": parsed_strata,
            "adjustments": parsed_adjustments,
        }
        parsed_flow_adjustments[param_name].append(parsed_adjustment)

    return parsed_flow_adjustments


def get_all_proportions(strata_names: List[str], strata_proportions: Dict[str, float]):
    """
    Determine what % of population get assigned to the different strata.
    """
    proportion_allocated = sum(strata_proportions.values())
    remaining_names = [n for n in strata_names if n not in strata_proportions]
    count_remaining = len(remaining_names)
    assert set(strata_proportions.keys()).issubset(strata_names), "Invalid proprotion keys"
    assert proportion_allocated <= 1.0000001, "Sum of proportions must not exceed 1.0"
    if not remaining_names:
        eps = 1e-12
        assert 1 - proportion_allocated < eps, "Sum of proportions must be 1.0"
        return strata_proportions
    else:
        # Divide the remaining proprotions equally
        starting_proportion = (1 - proportion_allocated) / count_remaining
        remaining_proportions = {name: starting_proportion for name in remaining_names}
        return {**strata_proportions, **remaining_proportions}


def get_stratified_compartment_names(
    strat: Stratification,
    comp_names: List[Compartment],
) -> List[Compartment]:
    """
    Stratify the model compartments into sub-compartments, based on the strata names provided,
    Only compartments specified in `comps_to_stratify` will be stratified.
    Returns the new compartment names.
    """
    new_comps = []
    for old_comp in comp_names:
        should_stratify = old_comp.has_name_in_list(strat.compartments)
        if should_stratify:
            for stratum in strat.strata:
                new_comp = old_comp.stratify(strat.name, stratum)
                new_comps.append(new_comp)
        else:
            new_comps.append(old_comp)

    return new_comps


def get_stratified_compartment_values(
    strat: Stratification,
    comp_names: List[Compartment],
    comp_values: np.ndarray,
) -> np.ndarray:
    """
    Stratify the model compartments into sub-compartments, based on the strata names provided,
    Split the population according to the provided proprotions.
    Only compartments specified in `comps_to_stratify` will be stratified.
    Returns the new compartment names and values
    """
    assert len(comp_names) == len(comp_values)
    new_comp_values = []
    for idx in range(len(comp_values)):
        should_stratify = comp_names[idx].has_name_in_list(strat.compartments)
        if should_stratify:
            for stratum in strat.strata:
                new_value = comp_values[idx] * strat.comp_split_props[stratum]
                new_comp_values.append(new_value)
        else:
            new_comp_values.append(comp_values[idx])

    return np.array(new_comp_values)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['summer',
 'summer.legacy',
 'summer.legacy.flow',
 'summer.legacy.flow.base',
 'summer.legacy.model',
 'summer.legacy.model.utils']

package_data = \
{'': ['*']}

install_requires = \
['Cerberus>=1.3.2,<2.0.0',
 'networkx==2.5',
 'numba>=0.52.0,<0.53.0',
 'numpy==1.19.5',
 'scipy==1.1.0']

setup_kwargs = {
    'name': 'summerepi',
    'version': '1.0.1',
    'description': 'Summer is a compartmental disease modelling framework, written in Python. It provides a high-level API to build and run models.',
    'long_description': '# Summer: compartmental disease modelling in Python\n\n[![Automated Tests](https://github.com/monash-emu/summer/actions/workflows/tests.yml/badge.svg)](https://github.com/monash-emu/summer/actions/workflows/tests.yml)\n\nSummer is a compartmental disease modelling framework, written in Python. It provides a high-level API to build and run models. Features include:\n\n- A variety of inter-compartmental flows (infections, sojourn, fractional, births, deaths, imports)\n- Force of infection multipliers (frequency, density)\n- Post-processing of compartment sizes into derived outputs\n- Stratification of compartments, including:\n  - Adjustments to flow rates based on strata\n  - Adjustments to infectiousness based on strata\n  - Heterogeneous mixing between strata\n  - Multiple disease strains\n\n**[Documentation here](http://summerepi.com/)** with [code examples](http://summerepi.com/examples).\n\n[Available on PyPi](https://pypi.org/project/summerepi/) as `summerepi`.\n\n## Installation and Quickstart\n\nThis project is tested with Python 3.6.\nInstall the `summerepi` package from PyPI\n\n```bash\npip install summerepi\n```\n\nThen you can use the library to build and run models. See here for some examples:\n\n- [SIR model](http://summerepi.com/examples/sir-model.html)\n- [SEIR model](http://summerepi.com/examples/seir-model.html)\n\nYou will find a significant performance improvement in the ODE solver if you set `OMP_NUM_THREADS` before importing `summer` or `numpy`.\n\n```python\n# Set this in your Python script\nos.environ["OMP_NUM_THREADS"] = "1"\n\n# Do it before importing summer or numpy\nimport summer\n# ...\n```\n\n## Development\n\n[Poetry](https://python-poetry.org/) is used for packaging and dependency management.\nYou will need to install poetry to work on this codebase.\nSome common things to do as a developer working on this codebase:\n\n```bash\n# Install requirements\npoetry config virtualenvs.in-project true\npoetry shell\npoetry install\n\n# Get a virtualenv for running other stuff\npoetry shell\n\n# Publish to PyPI - use your PyPI credentials\npoetry publish --build\n\n# Add a new package\npoetry add\n\n# Run tests\npoetry shell\npytest -vv\n\n# Format Python code\nblack .\n```\n\n## Releases\n\n_TODO_\n\n## Documentation\n\nSphinx is used to automatically build reference documentation for this library.\nThe documentation is automatically built and deployed to [summerepi.com](http://summerepi.com/) whenever code is pushed to `master`.\n\nTo build and deploy\n\n```bash\n./docs/scripts/build.sh\n./docs/scripts/deploy.sh\n```\n\nTo work on docs locally\n\n```bash\n# ./docs/scripts/watch.sh # FIXME - endless recursion\n./docs/scripts/build.sh\n# In a separate terminal\n./docs/scripts/serve.sh\n# Visit http://localhost:8000/\n```\n',
    'author': 'James Trauer',
    'author_email': 'james.trauer@monash.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'http://summerepi.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<3.9',
}


setup(**setup_kwargs)



"""
This module contains the main disease modelling class.
"""
import logging
import copy
from typing import Tuple, List, Dict, Callable, Optional
from functools import lru_cache

import networkx
import numpy as np
from scipy.interpolate import interp1d
from numba import jit

import summer.flows as flows
from summer.compartment import Compartment
from summer.stratification import Stratification
from summer.solver import solve_ode, SolverType
from summer.adjust import FlowParam

logger = logging.getLogger()

FlowRateFunction = Callable[[List[Compartment], np.ndarray, np.ndarray, float], np.ndarray]


class CompartmentalModel:
    """
    A compartmental disease model.

    This model defines a set of compartments which each contain a population.
    Disease dynamics are defined by a set of flows which link the compartments together.
    The model is run over a period of time, starting from some initial conditions to predict the future state of a disease.

    Args:
        times: The start and end times.
        compartments: The compartments to simulate.
        infectious_compartments: The compartments which are counted as infectious.
        time_step (optional): The timesteps to return results for. Does not affect the ODE solver. Defaults to ``1``.

    Attributes:
        times (np.ndarray): The times that the model will simulate.
        compartments (List[Compartment]): The model's compartments.
        initial_population (np.ndarray): The model's starting population. The indices of this
            array will match up with ``compartments``. This is zero by default and can be set with ``set_initial_population``.
        outputs (np.ndarray): The values of each compartment for each requested timestep. For ``C`` compartments and
            ``T`` timesteps this will be a ``CxT`` matrix. The column indices of this array will match up with ``compartments`` and the row indices will match up with ``times``.
        derived_outputs (Dict[str, np.ndarray]): Additional results that are caculated from ``outputs`` for each timestep.


    """

    _DEFAULT_DISEASE_STRAIN = "default"
    _DEFAULT_MIXING_MATRIX = np.array([[1]])

    def __init__(
        self,
        times: Tuple[int, int],
        compartments: List[str],
        infectious_compartments: List[str],
        timestep: float = 1.0,
    ):
        start_t, end_t = times
        assert start_t >= 0, "Start time must be >= 0"
        assert end_t > start_t, "End time must be greater than start time"
        start_t = round(start_t)
        end_t = round(end_t)
        time_period = end_t - start_t + 1
        num_steps = time_period / timestep
        assert num_steps >= 1, "Time step should be less than time period."
        assert num_steps % 1 == 0, "Time step should be a factor of time period"
        self.times = np.linspace(start_t, end_t, num=int(num_steps))

        msg = "Infectious compartments must be a subset of compartments"
        assert all(n in compartments for n in infectious_compartments), msg
        self.compartments = [Compartment(n) for n in compartments]
        self._infectious_compartments = [Compartment(n) for n in infectious_compartments]
        self.initial_population = np.zeros_like(self.compartments, dtype=np.float)

        # Keeps track of original, pre-stratified compartment names.
        self._original_compartment_names = [Compartment.deserialize(n) for n in compartments]
        # Tracks total deaths per timestep for death-replacement birth flows
        self._total_deaths = None
        # Keeps track of Stratifications that have been applied.
        self._stratifications = []
        # Flows to be applied to the model compartments
        self._flows = []
        # No outputs until the model has been run.
        self.outputs = None
        self.derived_outputs = None
        # Track derived output requests in a dictionary.
        self._derived_output_requests = {}
        # Track derived output request dependencies in a directed acylic graph (DAG).
        self._derived_output_graph = networkx.DiGraph()
        # Whitelist of derived outputs to evaluate
        self._derived_outputs_whitelist = []

        # Mixing matrices a list of NxN arrays used to calculate force of infection.
        self._mixing_matrices = []

        # A list of dicts that has the strata required to match a row in the mixing matrix.
        self._mixing_categories = [{}]
        # A list of the different sub-categories ('strains') of the diease that we are modelling.
        self._disease_strains = [self._DEFAULT_DISEASE_STRAIN]

        self._update_compartment_indices()

    def _update_compartment_indices(self):
        """
        Update the mapping of compartment name to idx for quicker lookups.
        """
        compartment_idx_lookup = {}
        # Update the mapping of compartment name to idx for quicker lookups.
        for idx, c in enumerate(self.compartments):
            c.idx = idx
            compartment_idx_lookup[c] = idx

        for flow in self._flows:
            flow.update_compartment_indices(compartment_idx_lookup)

    def set_initial_population(self, distribution: Dict[str, float]):
        """
        Sets the initial population of the model, which is zero by default.

        Args:
            distribution: A map of populations to be assigned to compartments.

        """
        error_msg = "Cannot set initial population after the model has been stratified"
        assert not self._stratifications, error_msg
        for idx, comp in enumerate(self.compartments):
            pop = distribution.get(comp.name, 0)
            assert pop >= 0, f"Population for {comp.name} cannot be negative: {pop}"
            self.initial_population[idx] = pop

    """
    Adding flows
    """

    def add_crude_birth_flow(
        self,
        name: str,
        birth_rate: FlowParam,
        dest: str,
        dest_strata: Optional[Dict[str, str]] = None,
        expected_flow_count: Optional[int] = None,
    ):
        """
        Adds a crude birth rate flow to the model.
        The number of births will be determined by the product of the birth rate and total population.

        Args:
            name: The name of the new flow.
            birth_rate: The fractional crude birth rate per timestep.
            dest: The name of the destination compartment.
            dest_strata (optional): A whitelist of strata to filter the destination compartments.
            expected_flow_count (optional): Used to assert that a particular number of flows are created.

        """
        self._validate_param(name, birth_rate)
        self._add_entry_flow(
            flows.CrudeBirthFlow,
            name,
            birth_rate,
            dest,
            dest_strata,
            expected_flow_count,
        )

    def add_replacement_birth_flow(
        self,
        name: str,
        dest: str,
        dest_strata: Optional[Dict[str, str]] = None,
        expected_flow_count: Optional[int] = None,
    ):
        """
        Adds a death-replacing birth flow to the model.
        The number of births will replace the total number of deaths each year

        Args:
            name: The name of the new flow.
            dest: The name of the destination compartment.
            dest_strata (optional): A whitelist of strata to filter the destination compartments.
            expected_flow_count (optional): Used to assert that a particular number of flows are created.

        """
        # Only allow a single replacement flow to be added to the model.
        is_already_used = any([type(f) is flows.ReplacementBirthFlow for f in self._flows])
        if is_already_used:
            msg = "There is already a replacement birth flow in this model, cannot add a second."
            raise ValueError(msg)

        self._add_entry_flow(
            flows.ReplacementBirthFlow,
            name,
            self._get_total_deaths,
            dest,
            dest_strata,
            expected_flow_count,
        )

    def add_importation_flow(
        self,
        name: str,
        num_imported: FlowParam,
        dest: str,
        dest_strata: Optional[Dict[str, str]] = None,
        expected_flow_count: Optional[int] = None,
    ):
        """
        Adds an importation flow to the model, where people enter the destination compartment from outside the system.
        The number of people imported per timestep is be completely determined by ``num_imported``.

        Args:
            name: The name of the new flow.
            num_imported: The number of people imported per timestep.
            dest: The name of the destination compartment.
            dest_strata (optional): A whitelist of strata to filter the destination compartments.
            expected_flow_count (optional): Used to assert that a particular number of flows are created.

        """
        self._validate_param(name, num_imported)
        self._add_entry_flow(
            flows.ImportFlow, name, num_imported, dest, dest_strata, expected_flow_count
        )

    def _add_entry_flow(
        self,
        flow_cls,
        name: str,
        param: FlowParam,
        dest: str,
        dest_strata: Optional[Dict[str, str]],
        expected_flow_count: Optional[int],
    ):
        dest_strata = dest_strata or {}
        dest_comps = [c for c in self.compartments if c.is_match(dest, dest_strata)]
        new_flows = []
        for dest_comp in dest_comps:
            flow = flow_cls(name, dest_comp, param)
            new_flows.append(flow)

        self._validate_expected_flow_count(expected_flow_count, new_flows)
        self._flows += new_flows
        self._update_compartment_indices()

    def add_death_flow(
        self,
        name: str,
        death_rate: FlowParam,
        source: str,
        source_strata: Optional[Dict[str, str]] = None,
        expected_flow_count: Optional[int] = None,
    ):
        """
        Adds a flow where people die and leave the compartment, reducing the total population.

        Args:
            name: The name of the new flow.
            death_rate: The fractional death rate per timestep.
            source: The name of the source compartment.
            source_strata (optional): A whitelist of strata to filter the source compartments.
            expected_flow_count (optional): Used to assert that a particular number of flows are created.

        """
        self._add_exit_flow(
            flows.DeathFlow,
            name,
            death_rate,
            source,
            source_strata,
            expected_flow_count,
        )

    def add_universal_death_flows(self, base_name: str, death_rate: FlowParam):
        """
        Adds a universal death rate flow to every compartment in the model.
        The number of deaths per compartment will be determined by the product of
        the death rate and the compartment population.

        The base name will be used to create the name of each flow. For example a
        base name of "universal_death" applied to the "S" comparement will result in a flow called
        "universal_death_for_S".

        Args:
            base_name: The base name for each new flow.
            death_rate: The fractional death rate per timestep.

        Returns:
            List[str]: The names of the flows added.

        """
        # Only allow a single universal death flow with a given name to be added to the model.
        is_already_used = any([f.name.startswith(base_name) for f in self._flows])
        if is_already_used:
            msg = f"There is already a universal death flow called '{base_name}' in this model, cannot add a second."
            raise ValueError(msg)

        flow_names = []
        for comp_name in self._original_compartment_names:
            flow_name = f"{base_name}_for_{comp_name}"
            flow_names.append(flow_name)
            self._add_exit_flow(
                flows.DeathFlow,
                flow_name,
                death_rate,
                comp_name,
                source_strata={},
                expected_flow_count=None,
            )

        return flow_names

    def _add_exit_flow(
        self,
        flow_cls,
        name: str,
        param: FlowParam,
        source: str,
        source_strata: Optional[Dict[str, str]],
        expected_flow_count: Optional[int],
    ):
        source_strata = source_strata or {}
        self._validate_param(name, param)
        source_comps = [c for c in self.compartments if c.is_match(source, source_strata)]
        new_flows = []
        for source_comp in source_comps:
            flow = flow_cls(name, source_comp, param)
            new_flows.append(flow)

        self._validate_expected_flow_count(expected_flow_count, new_flows)
        self._flows += new_flows
        self._update_compartment_indices()

    def add_sojourn_flow(
        self,
        name: str,
        sojourn_time: FlowParam,
        source: str,
        dest: str,
        source_strata: Optional[Dict[str, str]] = None,
        dest_strata: Optional[Dict[str, str]] = None,
        expected_flow_count: Optional[int] = None,
    ):
        """
        Adds a flow where that models a "sojourn" through a compartment, where the flow rate
        is proportional to the inverse of the sojourn time. For example if there is a sojourn time of 10
        days, then the flow rate will be 10% of the occupants per day.

        Args:
            name: The name of the new flow.
            sojourn_time: The mean time sojourn time for a person in the compartment.
            source: The name of the source compartment.
            dest: The name of the destination compartment.
            source_strata (optional): A whitelist of strata to filter the source compartments.
            dest_strata (optional): A whitelist of strata to filter the destination compartments.
            expected_flow_count (optional): Used to assert that a particular number of flows are created.

        """
        self._add_transition_flow(
            flows.SojournFlow,
            name,
            sojourn_time,
            source,
            dest,
            source_strata,
            dest_strata,
            expected_flow_count,
        )

    def add_infection_frequency_flow(
        self,
        name: str,
        contact_rate: FlowParam,
        source: str,
        dest: str,
        source_strata: Optional[Dict[str, str]] = None,
        dest_strata: Optional[Dict[str, str]] = None,
        expected_flow_count: Optional[int] = None,
    ):
        """
        Adds a flow that infects people using an "infection frequency" contact rate, which is
        when the infectious multiplier is determined by the proportion of infectious people to the total population.

        Args:
            name: The name of the new flow.
            contact_rate: The contact rate.
            source: The name of the source compartment.
            dest: The name of the destination compartment.
            source_strata (optional): A whitelist of strata to filter the source compartments.
            dest_strata (optional): A whitelist of strata to filter the destination compartments.
            expected_flow_count (optional): Used to assert that a particular number of flows are created.

        """
        self._add_transition_flow(
            flows.InfectionFrequencyFlow,
            name,
            contact_rate,
            source,
            dest,
            source_strata,
            dest_strata,
            expected_flow_count,
            find_infectious_multiplier=self._get_infection_frequency_multiplier,
        )

    def add_infection_density_flow(
        self,
        name: str,
        contact_rate: FlowParam,
        source: str,
        dest: str,
        source_strata: Optional[Dict[str, str]] = None,
        dest_strata: Optional[Dict[str, str]] = None,
        expected_flow_count: Optional[int] = None,
    ):
        """
        Adds a flow that infects people using an "infection density" contact rate, which is
        when the infectious multiplier is determined by the number of infectious people.

        Args:
            name: The name of the new flow.
            contact_rate: The contact rate.
            source: The name of the source compartment.
            dest: The name of the destination compartment.
            source_strata (optional): A whitelist of strata to filter the source compartments.
            dest_strata (optional): A whitelist of strata to filter the destination compartments.
            expected_flow_count (optional): Used to assert that a particular number of flows are created.

        """
        self._add_transition_flow(
            flows.InfectionDensityFlow,
            name,
            contact_rate,
            source,
            dest,
            source_strata,
            dest_strata,
            expected_flow_count,
            find_infectious_multiplier=self._get_infection_density_multiplier,
        )

    def add_fractional_flow(
        self,
        name: str,
        fractional_rate: FlowParam,
        source: str,
        dest: str,
        source_strata: Optional[Dict[str, str]] = None,
        dest_strata: Optional[Dict[str, str]] = None,
        expected_flow_count: Optional[int] = None,
    ):
        """
        Adds a flow transfers people from a source to a destination based on the population of the source
        compartment and the fractional flow rate.

        Args:
            name: The name of the new flow.
            fractional_rate: The fraction of people that transfer per timestep.
            source: The name of the source compartment.
            dest: The name of the destination compartment.
            source_strata (optional): A whitelist of strata to filter the source compartments.
            dest_strata (optional): A whitelist of strata to filter the destination compartments.
            expected_flow_count (optional): Used to assert that a particular number of flows are created.

        """
        self._add_transition_flow(
            flows.FractionalFlow,
            name,
            fractional_rate,
            source,
            dest,
            source_strata,
            dest_strata,
            expected_flow_count,
        )

    def add_function_flow(
        self,
        name: str,
        flow_rate_func: FlowRateFunction,
        source: str,
        dest: str,
        source_strata: Optional[Dict[str, str]] = None,
        dest_strata: Optional[Dict[str, str]] = None,
        expected_flow_count: Optional[int] = None,
    ):
        """
        A flow that transfers people from a source to a destination based on a user-defined function.
        This can be used to define more complex flows if required. See `flows.FunctionFlow` for more details on the arguments to the function.

        Args:
            name: The name of the new flow.
            flow_rate_func:  A function that returns the flow rate, before adjustments.
            source: The name of the source compartment.
            dest: The name of the destination compartment.
            source_strata (optional): A whitelist of strata to filter the source compartments.
            dest_strata (optional): A whitelist of strata to filter the destination compartments.
            expected_flow_count (optional): Used to assert that a particular number of flows are created.

        """
        self._add_transition_flow(
            flows.FunctionFlow,
            name,
            flow_rate_func,
            source,
            dest,
            source_strata,
            dest_strata,
            expected_flow_count,
        )

    def _add_transition_flow(
        self,
        flow_cls,
        name: str,
        param: FlowParam,
        source: str,
        dest: str,
        source_strata: Optional[Dict[str, str]],
        dest_strata: Optional[Dict[str, str]],
        expected_flow_count: Optional[int],
        find_infectious_multiplier=None,
    ):
        source_strata = source_strata or {}
        dest_strata = dest_strata or {}
        if flow_cls is not flows.FunctionFlow:
            # Non standard param for FunctionFlow so cannot use this validation method.
            self._validate_param(name, param)

        dest_comps = [c for c in self.compartments if c.is_match(dest, dest_strata)]
        source_comps = [c for c in self.compartments if c.is_match(source, source_strata)]
        num_dest = len(dest_comps)
        num_source = len(dest_comps)
        msg = f"Expected equal number of source and dest compartments, but got {num_source} source and {num_dest} dest."
        assert num_dest == num_source, msg
        new_flows = []
        for source_comp, dest_comp in zip(source_comps, dest_comps):
            if find_infectious_multiplier:
                flow = flow_cls(
                    name,
                    source_comp,
                    dest_comp,
                    param,
                    find_infectious_multiplier=find_infectious_multiplier,
                )
            else:
                flow = flow_cls(name, source_comp, dest_comp, param)

            new_flows.append(flow)

        self._validate_expected_flow_count(expected_flow_count, new_flows)
        self._flows += new_flows
        self._update_compartment_indices()

    def _validate_param(self, flow_name: str, param: FlowParam):
        """
        Ensure that the supplied parameter produces sensible results for all timesteps.
        """
        is_all_positive = (
            all(map(lambda t: param(t) >= 0, self.times)) if callable(param) else param >= 0
        )
        error_msg = f"Parameter for {flow_name} must be >= 0 for all timesteps: {param}"
        assert is_all_positive, error_msg

    def _validate_expected_flow_count(
        self, expected_count: Optional[int], new_flows: List[flows.BaseFlow]
    ):
        """
        Ensure the number of new flows created is the expected amount
        """
        if expected_count is not None:
            # Check that we added the expected number of flows.
            actual_count = len(new_flows)
            msg = f"Expected to add {expected_count} flows but added {actual_count}"
            assert actual_count == expected_count, msg

    """
    Stratifying the model
    """

    def stratify_with(self, strat: Stratification):
        """
        Apply the stratification to the model's flows and compartments.

        Args:
            strat: The stratification to apply.

        """
        # Validate flow adjustments
        flow_names = [f.name for f in self._flows]
        for n in strat.flow_adjustments.keys():
            msg = f"Flow adjustment for '{n}' refers to a flow that is not present in the model."
            assert n in flow_names, msg

        # Validate infectiousness adjustments.
        msg = "All stratification infectiousness adjustments must refer to a compartment that is present in model."
        assert all(
            [c in self._original_compartment_names for c in strat.infectiousness_adjustments.keys()]
        ), msg

        if strat.mixing_matrix is not None:
            # Add this stratification's mixing matrix if a new one is provided.
            assert not strat.is_strain(), "Strains cannot have a mixing matrix."
            # Only allow mixing matrix to be supplied if there is a complete stratification.
            msg = "Mixing matrices only allowed for full stratification."
            assert strat.compartments == self._original_compartment_names, msg
            self._mixing_matrices.append(strat.mixing_matrix)
            # Update mixing categories for force of infection calculation.
            old_mixing_categories = self._mixing_categories
            self._mixing_categories = []
            for mc in old_mixing_categories:
                for stratum in strat.strata:
                    self._mixing_categories.append({**mc, strat.name: stratum})

        if strat.is_strain():
            # Track disease strain names, overriding default values.
            msg = "A disease strain stratification has already been applied, cannot use more than one."
            assert not any([s.is_strain() for s in self._stratifications]), msg
            self._disease_strains = strat.strata

        # Stratify compartments, split according to split_proportions

        prev_compartment_names = copy.copy(self.compartments)
        self.compartments = strat._stratify_compartments(self.compartments)
        self.initial_population = strat._stratify_compartment_values(
            prev_compartment_names, self.initial_population
        )

        # Stratify flows
        prev_flows = self._flows
        self._flows = []
        for flow in prev_flows:
            self._flows += flow.stratify(strat)

        # Update the mapping of compartment name to idx for quicker lookups.
        self._update_compartment_indices()

        if strat.is_ageing():
            # Create inter-compartmental flows for ageing from one stratum to the next.
            # The ageing rate is proportional to the width of the age bracket.
            # It's assumed that both ages and model timesteps are in years.
            ages = list(sorted(map(int, strat.strata)))
            for age_idx in range(len(ages) - 1):
                start_age = int(ages[age_idx])
                end_age = int(ages[age_idx + 1])
                for comp in prev_compartment_names:
                    if not comp.has_name_in_list(strat.compartments):
                        # Don't include unstratified compartments
                        continue

                    source = comp.stratify(strat.name, str(start_age))
                    dest = comp.stratify(strat.name, str(end_age))
                    self.add_sojourn_flow(
                        name=f"ageing_{source}_to_{dest}",
                        sojourn_time=end_age - start_age,
                        source=source.name,
                        dest=dest.name,
                        source_strata=source.strata,
                        dest_strata=dest.strata,
                        expected_flow_count=1,
                    )

        self._stratifications.append(strat)

    """
    Running the model
    """

    def run(
        self,
        solver: str = SolverType.SOLVE_IVP,
        solver_args: Optional[dict] = None,
    ):
        """
        Runs the model over the provided time span, calculating the outputs and the derived outputs.
        The model calculates the outputs by solving an ODE which is defined by the initial population and the inter-compartmental flows.

        **Note**: The default ODE solver used to produce the model's outputs does not necessarily evaluate every requested timestep. This adaptive
        solver can skip over times, or double back when trying to characterize the ODE. The final results are produced by interpolating the
        solution produced by the ODE solver. This means that model dynamics that only occur in short time periods may not be reflected in the outputs.

        Args:
            solver (optional): The ODE solver to use, defaults to SciPy's IVP solver.
            solver_args (optional): Extra arguments to supplied to the solver, see ``summer.solver`` for details.

        """
        solver_args = solver_args or {}
        # Do some setup.
        self._prepare_to_run()
        # Calculate the outputs (compartment sizes) by solving the ODE defined by _get_flows().
        self.outputs = solve_ode(
            solver,
            self._get_flow_rates,
            self.initial_population,
            self.times,
            solver_args,
        )
        # Calculate any requested derived outputs, based on the calculated compartment sizes.
        self.derived_outputs = self._calculate_derived_outputs()

    def _prepare_to_run(self):
        """
        Pre-run setup.
        Here we do any calculations/preparation are possible to do before the model runs.
        """
        # Functions will often be called multiple times per timestep, so we cache any time-varying functions.
        # First we find all time varying functions.
        funcs = set()
        for flow in self._flows:
            if isinstance(flow, flows.FunctionFlow):
                # Don't cache these, the input arguments cannot be stored in a dict ("non hashable")
                continue
            elif callable(flow.param):
                funcs.add(flow.param)
            for adj in flow.adjustments:
                if adj and callable(adj.param):
                    funcs.add(adj.param)

        # Cache return values to prevent re-computation. This will a little leak memory, which is fine.
        funcs_cached = {}
        for func in funcs:
            # Floating point return type is 8 bytes, meaning 2**17 values is ~1MB of memory.
            funcs_cached[func] = lru_cache(maxsize=2 ** 17)(func)

        # Finally, replace original functions with cached ones
        for flow in self._flows:
            if flow.param in funcs_cached:
                flow.param = funcs_cached[flow.param]
            for adj in flow.adjustments:
                if adj and adj.param in funcs_cached:
                    adj.param = funcs_cached[adj.param]

        # Optimize flow adjustments
        for f in self._flows:
            f.optimize_adjustments()

        # Re-order flows so that they are executed in the correct order:
        #   - exit flows
        #   - entry flows (depends on exit flows for 'replace births' functionality)
        #   - transition flows
        #   - function flows (depend on all other prior flows, since they take flow rate as an input)
        num_flows = len(self._flows)
        _exit_flows = [f for f in self._flows if issubclass(f.__class__, flows.BaseExitFlow)]
        _entry_flows = [f for f in self._flows if issubclass(f.__class__, flows.BaseEntryFlow)]
        _transition_flows = [
            f
            for f in self._flows
            if issubclass(f.__class__, flows.BaseTransitionFlow)
            and not isinstance(f, flows.FunctionFlow)
        ]
        _function_flows = [f for f in self._flows if isinstance(f, flows.FunctionFlow)]
        self._has_function_flows = bool(_function_flows)
        self._flows = _exit_flows + _entry_flows + _transition_flows + _function_flows
        # Check we didn't miss any flows
        assert len(self._flows) == num_flows, "Some flows were lost when preparing to run."
        # Split flows into two groups for runtime.
        self._iter_function_flows = [
            (i, f) for i, f in enumerate(self._flows) if isinstance(f, flows.FunctionFlow)
        ]
        self._iter_non_function_flows = [
            (i, f) for i, f in enumerate(self._flows) if not isinstance(f, flows.FunctionFlow)
        ]

        # Prepare to track flow rates for derived outputs
        # We track the times and values of the request flows while integrating, and then
        # interpolate the results to get our flow-based derived outputs.
        # An ordered list of the times that have been tracked.
        self._flow_tracker_times = []
        # An ordered list of values for each flow, these will contain an array of flow rates for each timestep.
        self._flow_tracker_values = []

        """
        Pre-run calculations to help determine force of infection multiplier at runtime.

        We start with a set of "mixing categories". These categories describe groups of compartments.
        For example, we might have the stratifications age {child, adult} and location {work, home}.
        In this case, the mixing categories would be {child x home, child x work, adult x home, adult x work}.
        Mixing categories are only created when a mixing matrix is supplied during stratification.

        There is a mapping from every compartment to a mixing category.
        This is only true if mixing matrices are supplied only for complete stratifications.
        There are `num_cats` categories and `num_comps` compartments.
        The category matrix is a (num_cats x num_comps) matrix of 0s and 1s, with a 1 when the compartment is in a given category.
        We expect only one category per compartment, but there may be many compartments per category.

        We can multiply the category matrix by the vector of compartment sizes to get the total number of people
        in each mixing category.

        We also create a vector of values in [0, inf) which describes how infectious each compartment is: compartment_infectiousness
        We can use this vector plus the compartment sizes to get the 'effective' number of infectious people per compartment.

        We can use the 'effective infectious' compartment sizes, plus the mixing category matrix
        to get the infected population per mixing category.

        Now that we know:
            - the total population per category
            - the infected population per category
            - the inter-category mixing coefficients (mixing matrix)

        We can calculate the infection density or frequency per category.
        Finally, at runtime, we can lookup which category a given compartment is in and look up its infectious multiplier (density or frequency).
        """
        # Figure out which compartments should be infectious
        self._infectious_mask = [
            c.has_name_in_list(self._infectious_compartments) for c in self.compartments
        ]

        # Find out the relative infectiousness of each compartment, for each strain.
        # If no strains have been created, we assume a default strain name.
        self._compartment_infectiousness = {
            strain_name: self._get_compartment_infectiousness_for_strain(strain_name)
            for strain_name in self._disease_strains
        }

        # Create a matrix that tracks which categories each compartment is in.
        # A matrix with size (num_cats x num_comps).
        num_comps = len(self.compartments)
        num_categories = len(self._mixing_categories)
        self._category_lookup = {}  # Map compartments to categories.
        self._category_matrix = np.zeros((num_categories, num_comps))
        for i, category in enumerate(self._mixing_categories):
            for j, comp in enumerate(self.compartments):
                if all(comp.has_stratum(k, v) for k, v in category.items()):
                    self._category_matrix[i][j] = 1
                    self._category_lookup[j] = i

    def _get_compartment_infectiousness_for_strain(self, strain: str):
        """
        Returns a vector of floats, each representing the relative infectiousness of each compartment.
        If a strain name is provided, find the infectiousness factor *only for that strain*.
        """
        # Figure out which compartments should be infectious
        infectious_mask = np.array(
            [c.has_name_in_list(self._infectious_compartments) for c in self.compartments]
        )
        # Find the infectiousness multipliers for each compartment being implemented in the model.
        # Start from assumption that each compartment is not infectious.
        compartment_infectiousness = np.zeros(self.initial_population.shape)
        # Set all infectious compartments to be equally infectious.
        compartment_infectiousness[infectious_mask] = 1

        # Apply infectiousness adjustments.
        for idx, comp in enumerate(self.compartments):
            inf_value = compartment_infectiousness[idx]
            for strat in self._stratifications:
                for comp_name, adjustments in strat.infectiousness_adjustments.items():
                    for stratum, adjustment in adjustments.items():
                        should_apply_adjustment = adjustment and comp.is_match(
                            comp_name, {strat.name: stratum}
                        )
                        if should_apply_adjustment:
                            # Cannot use time-varying funtions for infectiousness adjustments,
                            # because this is calculated before the model starts running.
                            inf_value = adjustment.get_new_value(inf_value, None)

            compartment_infectiousness[idx] = inf_value

        if strain != self._DEFAULT_DISEASE_STRAIN:
            # Filter out all values that are not in the given strain.
            strain_mask = np.zeros(self.initial_population.shape)
            for idx, compartment in enumerate(self.compartments):
                if compartment.has_stratum("strain", strain):
                    strain_mask[idx] = 1

            compartment_infectiousness *= strain_mask

        return compartment_infectiousness

    def _prepare_time_step(self, time: float, compartment_values: np.ndarray):
        """
        Pre-timestep setup. This should be run before `_get_flow_rates`.
        Here we set up any stateful updates that need to happen before we get the flow rates.
        """
        # Prepare total deaths for tracking deaths.
        self._total_deaths = 0

        # Prepare derived output flow tracking for this timestep.
        self._flow_tracker_times.append(time)
        self._flow_tracker_values.append(np.zeros(len(self._flows)))

        # Find the effective infectious population for the force of infection calculations.
        mixing_matrix = self._get_mixing_matrix(time)

        # Calculate total number of people per category.
        # A vector with size (num_cats x 1).
        comp_values_transposed = compartment_values.reshape((compartment_values.shape[0], 1))
        self._category_populations = np.matmul(self._category_matrix, comp_values_transposed)

        # Calculate infectious populations for each strain.
        # Infection density / frequency is the infectious multiplier for each mixing category, calculated for each strain.
        self._infection_density = {}
        self._infection_frequency = {}
        for strain in self._disease_strains:
            strain_compartment_infectiousness = self._compartment_infectiousness[strain]

            # Calculate total infected people per category, including adjustment factors.
            # Returns a vector with size (num_cats x 1).
            infected_values = compartment_values * strain_compartment_infectiousness
            infected_values_transposed = infected_values.reshape((infected_values.shape[0], 1))
            infectious_populations = np.matmul(self._category_matrix, infected_values_transposed)
            self._infection_density[strain] = np.matmul(mixing_matrix, infectious_populations)

            # Calculate total infected person frequency per category, including adjustment factors.
            # A vector with size (num_cats x 1).
            category_frequency = infectious_populations / self._category_populations
            self._infection_frequency[strain] = np.matmul(mixing_matrix, category_frequency)

    def _get_force_idx(self, source: Compartment):
        """
        Returns the index of the source compartment in the infection multiplier vector.
        """
        return self._category_lookup[source.idx]

    def _get_total_deaths(self, *args, **kwargs) -> float:
        assert self._total_deaths is not None, "Total deaths has not been set."
        return self._total_deaths

    def _get_infection_frequency_multiplier(self, source: Compartment, dest: Compartment) -> float:
        """
        Get force of infection multiplier for a given compartment,
        using the 'infection frequency' calculation.
        """
        idx = self._get_force_idx(source)
        strain = dest.strata.get("strain", self._DEFAULT_DISEASE_STRAIN)
        return self._infection_frequency[strain][idx][0]

    def _get_infection_density_multiplier(self, source: Compartment, dest: Compartment):
        """
        Get force of infection multiplier for a given compartment,
        using the 'infection density' calculation.
        """
        idx = self._get_force_idx(source)
        strain = dest.strata.get("strain", self._DEFAULT_DISEASE_STRAIN)
        return self._infection_density[strain][idx][0]

    def _get_flow_rates(self, compartment_values: np.ndarray, time: float):
        """
        Get net flows into, out of, and between all compartments.
        This function is passed to solve_ode func and defines the dynamics of the model.
        """
        # Zero out -ve compartment sizes in flow rate calculations,
        # to prevent negative values from messing up the direction of flows.
        # We don't expect large -ve values, but there can be small ones due to numerical errors.
        comp_vals = compartment_values.copy()
        zero_mask = comp_vals < 0
        comp_vals[zero_mask] = 0
        self._prepare_time_step(time, comp_vals)
        flow_tracker_idx = len(self._flow_tracker_times) - 1
        # Track each flow's flow-rates at this point in time for derived outputs and function flows.
        flow_rates = self._flow_tracker_values[flow_tracker_idx]
        # Track the net flow rates between compartments for the ODE solver.
        net_flow_rates = np.zeros(len(comp_vals))

        # Find the flow rate for each flow.
        for flow_idx, flow in self._iter_non_function_flows:
            # Evaluate all the flows that are not function flows.
            net_flow = flow.get_net_flow(compartment_values, time)
            flow_rates[flow_idx] = net_flow
            if flow.source:
                net_flow_rates[flow.source.idx] -= net_flow
            if flow.dest:
                net_flow_rates[flow.dest.idx] += net_flow
            if flow.is_death_flow:
                # Track total deaths for any later birth replacement flows.
                self._total_deaths += net_flow

        if self._iter_function_flows:
            # Evaluate the function flows.
            original_flow_rates = flow_rates.copy()
            for flow_idx, flow in self._iter_function_flows:
                net_flow = flow.get_net_flow(
                    self.compartments, compartment_values, self._flows, original_flow_rates, time
                )
                flow_rates[flow_idx] = net_flow
                net_flow_rates[flow.source.idx] -= net_flow
                net_flow_rates[flow.dest.idx] += net_flow

        return net_flow_rates

    def _get_mixing_matrix(self, time: float) -> np.ndarray:
        """
        Returns the final mixing matrix for a given time.
        """
        mixing_matrix = self._DEFAULT_MIXING_MATRIX
        for mm_func in self._mixing_matrices:
            # Assume each mixing matrix is either an np.ndarray or a function of time that returns one.
            mm = mm_func(time) if callable(mm_func) else mm_func
            # Get Kronecker product of old and new mixing matrices.
            mixing_matrix = np.kron(mixing_matrix, mm)

        return mixing_matrix

    """
    Requesting and calculating derived outputs
    """
    _FLOW_REQUEST = "flow"
    _COMPARTMENT_REQUEST = "comp"
    _AGGREGATE_REQUEST = "agg"
    _CUMULATIVE_REQUEST = "cum"
    _FUNCTION_REQUEST = "func"

    def set_derived_outputs_whitelist(self, whitelist: List[str]):
        """
        Request that we should only calculate a subset of the model's derived outputs.
        This can be useful when you only care about some results and you want to cut down on runtime.

        Args:
            whitelist: A list of the derived output names to calculate, ignoring all others.

        """
        self._derived_outputs_whitelist = whitelist

    def _calculate_derived_outputs(self):
        """
        Calculates all requested derived outputs from the calculated compartment sizes.
        """
        assert self.outputs is not None, "Cannot calculate derived outputs: model has not been run."
        error_msg = "Cannot calculate derived outputs: dependency graph has cycles."
        assert networkx.is_directed_acyclic_graph(self._derived_output_graph), error_msg
        graph = self._derived_output_graph.copy()

        if self._derived_outputs_whitelist:
            # Only calculate the required outputs and their dependencies, ignore everything else.
            required_nodes = set()
            for name in self._derived_outputs_whitelist:
                # Find a list of the output required and its dependencies.
                output_dependencies = networkx.dfs_tree(graph.reverse(), source=name).reverse()
                required_nodes = required_nodes.union(output_dependencies.nodes)

            # Remove any nodes that aren't required from the graph.
            nodes = list(graph.nodes)
            for node in nodes:
                if not node in required_nodes:
                    graph.remove_node(node)

        derived_outputs = {}
        outputs_to_delete_after = []
        # Convert tracked flow values into a matrix where the 1st dimension is flow type, 2nd is time
        self._flow_tracker_values = np.array(self._flow_tracker_values).T

        # Calculate all the outputs in the correct order so that each output has its dependencies fulfilled.
        for name in networkx.topological_sort(graph):
            request = self._derived_output_requests[name]
            request_type = request["request_type"]
            output = np.zeros(self.times.shape)

            if not request["save_results"]:
                # Delete the results of this output once the calcs are done.
                outputs_to_delete_after.append(name)

            if request_type == self._FLOW_REQUEST:
                # User wants to track a set of flow rates over time.
                values = np.zeros(len(self._flow_tracker_times))
                for flow_idx, flow in enumerate(self._flows):
                    is_matching_flow = (
                        flow.name == request["flow_name"]
                        and ((not flow.source) or flow.source.has_strata(request["source_strata"]))
                        and ((not flow.dest) or flow.dest.has_strata(request["dest_strata"]))
                    )
                    if is_matching_flow:
                        values += self._flow_tracker_values[flow_idx]

                # Build a function to produce interpolated results
                solved_func = interp1d(self._flow_tracker_times, values)
                # Populate output array with interpolated results
                num_times = len(self.times)
                interpolated_output = np.zeros(self.times.shape)
                for time_idx in range(num_times):
                    interpolated_output[time_idx] = solved_func(self.times[time_idx])

                use_raw_results = request["raw_results"]
                if use_raw_results:
                    # Use interpolated flow rates wiuth no post-processing.
                    output = interpolated_output
                else:
                    # Set the "flow rate" at time `t` to be an estimate of the flow rate
                    # that is calculated at time `t-1`. By convention, flows are zero at t=0.
                    # This is done so that we can estimate the number of people moving between compartments
                    # using tracked flow rates.
                    ignore_first_timestep_output = np.zeros(self.times.shape)
                    ignore_first_timestep_output[1:] = interpolated_output[1:]
                    offset_output = np.zeros(self.times.shape)
                    offset_output[1:] = interpolated_output[:-1]
                    output = (offset_output + ignore_first_timestep_output) / 2

            elif request_type == self._COMPARTMENT_REQUEST:
                # User wants to track a set of compartment sizes over time.
                compartments = request["compartments"]
                strata = request["strata"]
                comps = (
                    (i, c)
                    for i, c in enumerate(self.compartments)
                    if c.has_name_in_list(compartments)
                )
                idxs = [i for i, c in comps if c.is_match(c.name, strata)]
                output = self.outputs[:, idxs].sum(axis=1)

            elif request_type == self._AGGREGATE_REQUEST:
                # User wants to track the sum of a set of outputs over time.
                source_names = request["sources"]
                output = sum([derived_outputs[s] for s in source_names])

            elif request_type == self._CUMULATIVE_REQUEST:
                # User wants to track cumulative value of an output over time.
                source_name = request["source"]
                start_time = request["start_time"]
                max_time = self.times.max()
                if start_time and start_time > max_time:
                    # Handle case where the derived output starts accumulating after the last model timestep.
                    msg = f"Cumulative output '{name}' start time {start_time} is greater than max model time {max_time}, defaulting to {max_time}"
                    logger.warn(msg)
                    start_time = max_time

                if start_time is None:
                    output = np.cumsum(derived_outputs[source_name])
                else:
                    assert (
                        start_time in self.times
                    ), f"Start time {start_time} not in times for '{name}'"
                    start_idx = np.where(self.times == start_time)[0][0]
                    output[start_idx:] = np.cumsum(derived_outputs[source_name][start_idx:])

            elif request_type == self._FUNCTION_REQUEST:
                # User wants to track the results of a function of other outputs over time.
                func = request["func"]
                source_names = request["sources"]
                inputs = [derived_outputs[s] for s in source_names]
                output = func(*inputs)

            derived_outputs[name] = output

        # Delete any intermediate outputs that we don't want to save.
        for name in outputs_to_delete_after:
            del derived_outputs[name]

        return derived_outputs

    def request_output_for_flow(
        self,
        name: str,
        flow_name: str,
        source_strata: Optional[Dict[str, str]] = None,
        dest_strata: Optional[Dict[str, str]] = None,
        save_results: bool = True,
        raw_results: bool = False,
    ):
        """
        Adds a derived output to the model's results. The output
        will be the value of the requested flow at the at each timestep.

        Args:
            name: The name of the derived output.
            flow_name: The name of the flow to track.
            source_strata (optional): A whitelist of strata to filter the source compartments.
            dest_strata (optional): A whitelist of strata to filter the destination compartments.
            save_results (optional): Whether to save or discard the results. Defaults to ``True``.
            raw_results (optional): Whether to use raw interpolated flow rates, or post-process them so that they're more
                represenative of the changes in compartment sizes. Defaults to ``False``.
        """
        source_strata = source_strata or {}
        dest_strata = dest_strata or {}
        msg = f"A derived output named {name} already exists."
        assert name not in self._derived_output_requests, msg
        is_flow_exists = any(
            [f.is_match(flow_name, source_strata, dest_strata) for f in self._flows]
        )
        assert is_flow_exists, f"No flow matches: {flow_name} {source_strata} {dest_strata}"
        self._derived_output_graph.add_node(name)
        self._derived_output_requests[name] = {
            "request_type": self._FLOW_REQUEST,
            "flow_name": flow_name,
            "source_strata": source_strata,
            "dest_strata": dest_strata,
            "raw_results": raw_results,
            "save_results": save_results,
        }

    def request_output_for_compartments(
        self,
        name: str,
        compartments: List[str],
        strata: Optional[Dict[str, str]] = None,
        save_results: bool = True,
    ):
        """
        Adds a derived output to the model's results. The output
        will be the aggregate population of the requested compartments at the at each timestep.

        Args:
            name: The name of the derived output.
            compartments: The name of the compartments to track.
            strata (optional): A whitelist of strata to filter the compartments.
            save_results (optional): Whether to save or discard the results.
        """
        strata = strata or {}
        msg = f"A derived output named {name} already exists."
        assert name not in self._derived_output_requests, msg
        is_match_exists = any(
            [any([c.is_match(name, strata) for name in compartments]) for c in self.compartments]
        )
        assert is_match_exists, f"No compartment matches: {compartments} {strata}"
        self._derived_output_graph.add_node(name)
        self._derived_output_requests[name] = {
            "request_type": self._COMPARTMENT_REQUEST,
            "compartments": compartments,
            "strata": strata,
            "save_results": save_results,
        }

    def request_aggregate_output(
        self,
        name: str,
        sources: List[str],
        save_results: bool = True,
    ):
        """
        Adds a derived output to the model's results. The output will be the aggregate of other derived outputs.

        Args:
            name: The name of the derived output.
            sources: The names of the derived outputs to aggregate.
            save_results (optional): Whether to save or discard the results.

        """
        msg = f"A derived output named {name} already exists."
        assert name not in self._derived_output_requests, msg
        for source in sources:
            assert (
                source in self._derived_output_requests
            ), f"Source {source} has not been requested."
            self._derived_output_graph.add_edge(source, name)

        self._derived_output_graph.add_node(name)
        self._derived_output_requests[name] = {
            "request_type": self._AGGREGATE_REQUEST,
            "sources": sources,
            "save_results": save_results,
        }

    def request_cumulative_output(
        self,
        name: str,
        source: str,
        start_time: int = None,
        save_results: bool = True,
    ):
        """
        Adds a derived output to the model's results. The output will be the cumulative value
        of another derived outputs over the model's time period.

        Args:
            name: The name of the derived output.
            source: The name of the derived outputs to accumulate.
            start_time (optional): The time to start accumulating from, defaults to model start time.
            save_results (optional): Whether to save or discard the results.

        """
        msg = f"A derived output named {name} already exists."
        assert name not in self._derived_output_requests, msg
        assert source in self._derived_output_requests, f"Source {source} has not been requested."
        self._derived_output_graph.add_node(name)
        self._derived_output_graph.add_edge(source, name)
        self._derived_output_requests[name] = {
            "request_type": self._CUMULATIVE_REQUEST,
            "source": source,
            "start_time": start_time,
            "save_results": save_results,
        }

    def request_function_output(
        self,
        name: str,
        func: Callable[[np.ndarray], np.ndarray],
        sources: List[str],
        save_results: bool = True,
    ):
        """
        Adds a derived output to the model's results. The output will be the result of a function
        which takes a list of sources as an input.

        Args:
            name: The name of the derived output.
            func: A function used to calculate the derived ouput.
            sources: The derived ouputs to input into the function.
            save_results (optional): Whether to save or discard the results.

        Example:
            Request a function-based derived ouput::

                model.request_output_for_compartments(
                    compartments=["S", "E", "I", "R"],
                    name="total_population",
                    save_results=False
                )
                model.request_output_for_compartments(
                    compartments=["R"],
                    name="recovered_population",
                     save_results=False
                )

                def calculate_proportion_seropositive(recovered_pop, total_pop):
                    return recovered_pop / total_pop

                model.request_function_output(
                    name="proportion_seropositive",
                    func=calculate_proportion_seropositive,
                    sources=["recovered_population", "total_population"],
                )

        """
        msg = f"A derived output named {name} already exists."
        assert name not in self._derived_output_requests, msg
        for source in sources:
            assert (
                source in self._derived_output_requests
            ), f"Source {source} has not been requested."
            self._derived_output_graph.add_edge(source, name)

        self._derived_output_graph.add_node(name)
        self._derived_output_requests[name] = {
            "request_type": self._FUNCTION_REQUEST,
            "func": func,
            "sources": sources,
            "save_results": save_results,
        }



from typing import List, Dict, Callable

import numpy as np
from numpy.lib.function_base import append

from summer.legacy.constants import Flow


class DerivedOutputCalculator:
    def __init__(self):
        """
        Calculates derived outputs from a set of outputs
        """
        self._flow_outputs = {}
        self._function_outputs = []

    def add_flow_derived_outputs(self, flow_outputs: dict):
        self._flow_outputs.update(flow_outputs)

    def add_function_derived_outputs(self, func_outputs: dict):
        for k, v in func_outputs.items():
            self._function_outputs.append([k, v])

    def calculate(self, model):
        """
        Returns the derived outputs for a model.
        """
        derived_outputs = {}

        # Initialize outputs
        base_output = np.zeros(model.times.shape)

        # Initialize flow outputs
        flow_lookup = {}
        for output_name, flow_output in self._flow_outputs.items():
            flow_lookup[output_name] = flow_output.filter_flows(model.flows)
            derived_outputs[output_name] = base_output.copy()

        # Initialize function outputs
        for output, _ in self._function_outputs:
            derived_outputs[output] = base_output.copy()

        # Calculate derived outputs for all model time steps.
        for time_idx, time in enumerate(model.times):
            compartment_values = model.restore_past_state(time_idx)
            # Calculate derived flow outputs.
            for output_name, output_flows in flow_lookup.items():
                for flow in output_flows:
                    net_flow = flow.get_net_flow(compartment_values, time)
                    derived_outputs[output_name][time_idx] += net_flow

            # Calculate derived function outputs
            for output_name, func in self._function_outputs:
                derived_outputs[output_name][time_idx] = func(
                    time_idx, model, compartment_values, derived_outputs
                )

        return derived_outputs


class InfectionDeathFlowOutput:
    """
    A request to track a set of infection death flow in the model's derived outputs.
    """

    def __init__(
        self,
        source: str,
        source_strata: Dict[str, str],
    ):
        self.source = source
        self.source_strata = source_strata

    def filter_flows(self, flows):
        filtered_flows = []
        for flow in flows:
            if flow.type != Flow.DEATH:
                continue

            if not flow.source.has_name(self.source):
                continue

            # Check source strata
            if self.source_strata:
                is_source_strata_valid = all(
                    flow.source.has_stratum(k, v) for k, v in self.source_strata.items()
                )
                if not is_source_strata_valid:
                    continue

            filtered_flows.append(flow)

        return filtered_flows


class TransitionFlowOutput:
    """
    A request to track a set of transition flows in the model's derived outputs.
    """

    def __init__(
        self,
        source: str,
        dest: str,
        source_strata: Dict[str, str],
        dest_strata: Dict[str, str],
    ):
        self.source = source
        self.dest = dest
        self.source_strata = source_strata
        self.dest_strata = dest_strata

    def filter_flows(self, flows):
        filtered_flows = []
        for flow in flows:
            if flow.type not in Flow.TRANSITION_FLOWS:
                continue

            if not flow.source.has_name(self.source):
                continue

            if not flow.dest.has_name(self.dest):
                continue

            # Check source strata
            if self.source_strata:
                is_source_strata_valid = all(
                    flow.source.has_stratum(k, v) for k, v in self.source_strata.items()
                )
                if not is_source_strata_valid:
                    continue

            # Check dest strata
            if self.dest_strata:
                is_dest_strata_valid = all(
                    flow.dest.has_stratum(k, v) for k, v in self.dest_strata.items()
                )
                if not is_dest_strata_valid:
                    continue

            filtered_flows.append(flow)

        return filtered_flows



import logging
from typing import List, Dict
from functools import lru_cache

import numpy as np

from summer.legacy.constants import (
    Flow,
    BirthApproach,
    IntegrationType,
)
from summer.legacy.compartment import Compartment
from summer.legacy.flow import (
    CrudeBirthFlow,
    InfectionDensityFlow,
    InfectionFrequencyFlow,
    ReplacementBirthFlow,
    StandardFlow,
    InfectionDeathFlow,
    UniversalDeathFlow,
    ImportFlow,
)

from .utils.solver import solve_ode
from .utils.validation import validate_model
from .derived_outputs import DerivedOutputCalculator


logger = logging.getLogger(__name__)

OUTPUTS_NEGATIVE_TOLERANCE = -1e0


class EpiModel:
    """
    Constructs and runs compartment-based epidemiological models,
    typically of infectious disease transmission.

    Compartment attributes
    compartment_names: The model compartments.
    compartment_values: The compartment sizes.
    entry_compartment: The compartment to add new population members to.
    initial_conditions: Proprotion of starting population for each compartment.
    starting_population: Total starting population.

    Flow attributes
    flows:  The flows between compartments
    parameters: Weights for each flow.
    time_variants: Time-varying flow weights.
    birth_approach: How we handle births.

    Runtime attributes
    times: Time steps at which outputs are to be evaluated.
    outputs: All the evaluated compartment sizes over requested times.
    total_deaths: Tracks total deaths for the 'replace deaths' birth approach.
    infectious_compartment: Names of the infectious compartments for infection flows
    populations_infectious: The current size of the infectious population.
    total_population: The current size of the total population, relevant for frequency-dependent transmission.
    infectious_mask: Numpy mask for identifying the infectious compartments.
    """

    def __init__(
        self,
        times: List[float],
        compartment_names: List[str],
        initial_conditions: Dict[str, float],
        parameters: Dict[str, float],
        requested_flows: List[dict],
        birth_approach: str,
        starting_population: int,
        infectious_compartments: List[str],
        entry_compartment: str,
    ):
        model_kwargs = {
            "times": times,
            "compartment_names": compartment_names,
            "initial_conditions": initial_conditions,
            "parameters": parameters,
            "requested_flows": requested_flows,
            "birth_approach": birth_approach,
            "starting_population": starting_population,
            "infectious_compartments": infectious_compartments,
            "entry_compartment": entry_compartment,
        }
        validate_model(model_kwargs)
        # Flows
        self.flows = []
        self.time_variants = {}
        self.parameters = {}

        # Compartments
        self.compartment_names = [Compartment(n) for n in compartment_names]
        self.infectious_compartments = [Compartment(n) for n in infectious_compartments]

        # Derived outputs
        self.derived_outputs = {}
        self._derived_calc = DerivedOutputCalculator()

        self.times = times

        # Populate model compartments with the values set in `initial conditions`.
        self.compartment_values = np.zeros(len(compartment_names))
        pop_remainder = starting_population - sum(initial_conditions.values())
        for idx, comp_name in enumerate(compartment_names):
            if comp_name in initial_conditions:
                self.compartment_values[idx] = initial_conditions[comp_name]

            if comp_name == entry_compartment:
                self.compartment_values[idx] += pop_remainder

        self.birth_approach = birth_approach
        self.parameters = parameters
        self.parameters["crude_birth_rate"] = self.parameters.get("crude_birth_rate", 0)
        self.parameters["universal_death_rate"] = self.parameters.get("universal_death_rate", 0)

        # Add user-specified flows
        self.entry_compartment = Compartment.deserialize(entry_compartment)
        self.flows = []
        for requested_flow in requested_flows:
            self._add_flow(requested_flow)

        # Add birth flows.
        if (
            self.birth_approach == BirthApproach.ADD_CRUDE
            and self.parameters["crude_birth_rate"] > 0
        ):
            flow = CrudeBirthFlow(
                dest=self.entry_compartment,
                param_name="crude_birth_rate",
                param_func=self.get_parameter_value,
            )
            self.flows.append(flow)
        elif self.birth_approach == BirthApproach.REPLACE_DEATHS:
            self.total_deaths = 0
            flow = ReplacementBirthFlow(
                dest=self.entry_compartment, get_total_deaths=self.get_total_deaths
            )
            self.flows.append(flow)

        # Add non-disease death flows if requested
        if self.parameters["universal_death_rate"] > 0:
            for compartment_name in self.compartment_names:
                flow = UniversalDeathFlow(
                    source=compartment_name,
                    param_name="universal_death_rate",
                    param_func=self.get_parameter_value,
                )
                self.flows.append(flow)

        # Update indices for quick lookups.
        self._update_flow_compartment_indices()
        for idx, c in enumerate(self.compartment_names):
            c.idx = idx

    def _update_flow_compartment_indices(self):
        """
        Create lookup table for quickly getting / setting compartments by name
        """
        compartment_idx_lookup = {name: idx for idx, name in enumerate(self.compartment_names)}
        for flow in self.flows:
            flow.update_compartment_indices(compartment_idx_lookup)

    def _add_flow(self, requested_flow):
        flow = None
        f_source = requested_flow.get("origin")
        f_source = Compartment.deserialize(f_source) if f_source else None
        f_dest = requested_flow.get("to")
        f_dest = Compartment.deserialize(f_dest) if f_dest else None
        f_param = requested_flow["parameter"]
        f_type = requested_flow["type"]
        if f_type == Flow.STANDARD:
            flow = StandardFlow(
                source=f_source,
                dest=f_dest,
                param_name=f_param,
                param_func=self.get_parameter_value,
            )
        elif f_type == Flow.INFECTION_FREQUENCY:
            flow = InfectionFrequencyFlow(
                source=f_source,
                dest=f_dest,
                param_name=f_param,
                param_func=self.get_parameter_value,
                find_infectious_multiplier=self.get_infection_frequency_multiplier,
            )
        elif f_type == Flow.INFECTION_DENSITY:
            flow = InfectionDensityFlow(
                source=f_source,
                dest=f_dest,
                param_name=f_param,
                param_func=self.get_parameter_value,
                find_infectious_multiplier=self.get_infection_density_multiplier,
            )
        elif f_type == Flow.DEATH:
            flow = InfectionDeathFlow(
                source=f_source, param_name=f_param, param_func=self.get_parameter_value
            )
        elif f_type == Flow.IMPORT:
            flow = ImportFlow(
                dest=self.entry_compartment,
                param_name=f_param,
                param_func=self.get_parameter_value,
            )

        if flow:
            self.flows.append(flow)

        return flow

    # Cache return values to prevent re-computation. This will a little leak memory, which is fine.
    # Floating point return type is 8 bytes, meaning 2**17 values is ~1MB of memory.
    @lru_cache(maxsize=2 ** 17)
    def get_parameter_value(self, name: str, time: float):
        """
        Get parameter value at a given time
        """
        if name in self.time_variants:
            return self.time_variants[name](time)
        else:
            return self.parameters[name]

    def prepare_to_run(self):
        """
        Pre-run setup.
        """
        # Split flows up into 3 groups: entry, exit and transition.
        self.entry_flows = [f for f in self.flows if f.type in Flow.ENTRY_FLOWS]
        self.exit_flows = [f for f in self.flows if f.type in Flow.EXIT_FLOWS]
        self.transition_flows = [f for f in self.flows if f.type in Flow.TRANSITION_FLOWS]

        # Check we didn't miss any flows
        split_flows = [self.entry_flows, self.exit_flows, self.transition_flows]
        num_split_flows = sum(len(fs) for fs in split_flows)
        assert len(self.flows) == num_split_flows

        # Determine order of how flows are run.
        # We apply deaths before births so that we can use 'total deaths' to calculate the birth rate, if required.
        self.flow_functions = [
            self.apply_transition_flows,
            self.apply_exit_flows,
            self.apply_entry_flows,
        ]
        self.prepare_force_of_infection()

    def prepare_force_of_infection(self):
        """
        Pre-run calculations to help determine force of infection multiplier at runtime.
        """
        # Figure out which compartments should be infectious
        self.infectious_mask = [
            c.has_name_in_list(self.infectious_compartments) for c in self.compartment_names
        ]

    def run_model(self, integration_type=IntegrationType.SOLVE_IVP, solver_args={}):
        """
        Calculates the model's outputs using an ODE solver.

        The ODE is solved over the user-specified timesteps (self.times),
        using the user-specified initial conditions (self.compartment_values).

        The final result is an array of compartment values at each timestep (self.outputs).
        Also calculates post-processing outputs after the ODE integration is complete.
        """
        self.prepare_to_run()
        self.outputs = solve_ode(
            integration_type, self.get_flow_rates, self.compartment_values, self.times, solver_args
        )

        self.derived_outputs = self.calculate_derived_outputs()

    def get_flow_rates(self, compartment_values: np.ndarray, time: float):
        """
        Get net flows into, out of, and between all compartments.
        Order of args determined by solve_ode func.
        """
        # Zero out compartment sizes to prevent negative values from messing up calcs.
        comp_vals = compartment_values.copy()
        zero_mask = comp_vals < 0
        comp_vals[zero_mask] = 0
        self.prepare_time_step(time, comp_vals)
        flow_rates = np.zeros(comp_vals.shape)
        for flow_func in self.flow_functions:
            flow_rates = flow_func(flow_rates, comp_vals, time)

        return flow_rates

    def prepare_time_step(self, time: float, compartment_values: np.ndarray):
        self.total_deaths = 0
        self.find_infectious_population(time, compartment_values)

    def apply_transition_flows(
        self, flow_rates: np.ndarray, compartment_values: np.ndarray, time: float
    ):
        """
        Apply fixed or infection-related inter-compartmental transition flows.
        """
        for flow in self.transition_flows:
            net_flow = flow.get_net_flow(compartment_values, time)
            flow_rates[flow.source.idx] -= net_flow
            flow_rates[flow.dest.idx] += net_flow

        return flow_rates

    def apply_exit_flows(self, flow_rates: np.ndarray, compartment_values: np.ndarray, time: float):
        """
        Apply exit flows: deaths, exports, etc.
        """
        for flow in self.exit_flows:
            net_flow = flow.get_net_flow(compartment_values, time)
            flow_rates[flow.source.idx] -= net_flow
            if flow.type in Flow.DEATH_FLOWS:
                self.total_deaths += net_flow

        return flow_rates

    def apply_entry_flows(
        self, flow_rates: np.ndarray, compartment_values: np.ndarray, time: float
    ):
        """
        Apply entry flows: births, imports, etc.
        """
        for flow in self.entry_flows:

            net_flow = flow.get_net_flow(compartment_values, time)
            flow_rates[flow.dest.idx] += net_flow

        return flow_rates

    def get_total_deaths(self):
        return self.total_deaths

    def get_infection_frequency_multiplier(self, source: Compartment, dest: Compartment):
        return self.population_infectious / self.population_total

    def get_infection_density_multiplier(self, source: Compartment, dest: Compartment):
        return self.population_infectious

    def find_infectious_population(self, time: float, compartment_values: np.ndarray):
        """
        Finds the effective infectious population
        """
        self.population_infectious = sum(compartment_values[self.infectious_mask])
        self.population_total = sum(compartment_values)

    def restore_past_state(self, time_idx: int):
        """
        Update the model's tracked quantities to a particular time step.
        This is used for calculating derived outputs.
        Returns the compartment values for that time step.
        """
        time = self.times[time_idx]
        compartment_values = self.outputs[time_idx]
        self.prepare_time_step(time, compartment_values)
        return compartment_values

    def calculate_derived_outputs(self):
        return self._derived_calc.calculate(self)

    def add_flow_derived_outputs(self, flow_outputs: dict):
        self._derived_calc.add_flow_derived_outputs(flow_outputs)

    def add_function_derived_outputs(self, func_outputs):
        self._derived_calc.add_function_derived_outputs(func_outputs)
